
"""
Contains system specifc information. At the Moment, this is all hard-coded.
In the future, this will be part of the installation of AIzymes.

set_system() contains general variables.
submit_head() constructs the submission header to submit jobs.

"""

import sys
import os
import getpass

def set_system(self):
    
    if self.SYSTEM == 'GRID':       
        
        self.rosetta_ext          = "linuxgccrelease"
        self.bash_args            = ""
        self.ROSETTA_PATH         = "/home/bunzelh/rosetta_src_2021.16.61629_bundle/main/source/"
        self.FIELD_TOOLS          = f'{self.FOLDER_HOME}/../../../src/FieldTools.py'
        self.FOLDER_ProteinMPNN   = f'{os.path.expanduser("~")}/ProteinMPNN'
        self.FOLDER_ProteinMPNN_h = f'{os.path.expanduser("~")}/ProteinMPNN/helper_scripts'
            
    elif self.SYSTEM == 'GRID_parallel':       
        
        self.rosetta_ext          = "linuxgccrelease"
        self.bash_args            = ""
        self.ROSETTA_PATH         = "/home/bunzelh/rosetta_src_2021.16.61629_bundle/main/source/"
        self.FIELD_TOOLS          = f'{self.FOLDER_HOME}/../../../src/FieldTools.py'
        self.FOLDER_ProteinMPNN   = f'{os.path.expanduser("~")}/ProteinMPNN'
        self.FOLDER_ProteinMPNN_h = f'{os.path.expanduser("~")}/ProteinMPNN/helper_scripts'
        
    elif self.SYSTEM == 'SCC':       
        
        self.rosetta_ext          = "linuxgccrelease"
        self.bash_args            = ""
        self.ROSETTA_PATH         = "/scratch-scc/projects/scc_mmtm_bunzel/rosetta.source.release-371/main/source/"
        self.FIELD_TOOLS          = f'{self.FOLDER_HOME}/../../../src/FieldTools.py'
        self.FOLDER_ProteinMPNN   = f'{os.path.expanduser("~")}/ProteinMPNN'
        self.FOLDER_ProteinMPNN_h = f'{os.path.expanduser("~")}/ProteinMPNN/helper_scripts'  
        
    elif self.SYSTEM == 'RAVEN':       
        
        self.rosetta_ext          = "linuxgccrelease"
        self.bash_args            = ""
        self.ROSETTA_PATH         = f"/u/bunzela/bin/rosetta.source.release-371/main/source/"
        self.FIELD_TOOLS          = f'/u/{getpass.getuser()}/bin/AIzymes_dev/src/FieldTools.py'
        self.FOLDER_ProteinMPNN   = f'/u/{getpass.getuser()}/bin/ProteinMPNN'
        self.FOLDER_ProteinMPNN_h = f'/u/{getpass.getuser()}/bin/ProteinMPNN/helper_scripts'
        self.FOLDER_LigandMPNN    = f'/u/{getpass.getuser()}/bin/LigandMPNN'
                
    elif self.SYSTEM == 'BLUEPEBBLE':     
        
        self.rosetta_ext          = "linuxgccrelease"
        self.bash_args            = ""
        self.ROSETTA_PATH         = "/user/work/qz22231/rosetta.source.release-371/main/source"
        self.FIELD_TOOLS          = "/user/work/qz22231/AIzymes/AIzymes_git/AIzymes/src/FieldTools.py" #'../../../src/FieldTools.py'
        self.FOLDER_ProteinMPNN   = "/user/work/qz22231/ProteinMPNN"
        self.FOLDER_ProteinMPNN_h = "/user/work/qz22231/ProteinMPNN/helper_scripts"
        self.BLUEPEBBLE_ACCOUNT   = "ptch000721"
                
    elif self.SYSTEM == 'ABBIE_LOCAL':    
        
        self.rosetta_ext          = "linuxgccrelease"
        self.bash_args            = "OMP_NUM_THREADS=1 "
        self.ROSETTA_PATH         = "ERROR! NOT YET CONFIGURED"
        self.FIELD_TOOLS          = f'{self.FOLDER_HOME}/../../../src/FieldTools.py'
        self.SUBMIT_HEAD          = ""
        self.FOLDER_ProteinMPNN   = f'{os.path.expanduser("~")}/ProteinMPNN'
        self.FOLDER_ProteinMPNN_h = f'{os.path.expanduser("~")}/ProteinMPNN/helper_scripts'
        
    else:

        print(f"{self.SYSTEM} not defined in set_system(self), located in setup_system.py.")
        sys.exit()
        
def submit_head(self, index, job, ram):

    # Note: All RUN_INTERACTIVE jobs are also RUN_PARALLEL jobs!
    if self.RUN_PARALLEL == True:
        
        return ""
            
    elif self.SYSTEM == 'GRID': 

        return f"""#!/bin/bash
#$ -V
#$ -cwd
#$ -N {self.SUBMIT_PREFIX}_{job}_{index}
#$ -hard -l mf={ram}G
#$ -o {self.FOLDER_HOME}/{index}/scripts/{job}_{index}.out
#$ -e {self.FOLDER_HOME}/{index}/scripts/{job}_{index}.err
"""         
          
    elif self.SYSTEM == 'BLUEPEBBLE':     
        
        return f"""#!/bin/bash
#SBATCH --account={self.BLUEPEBBLE_ACCOUNT}
#SBATCH --partition=short
#SBATCH --mem=40GB
#SBATCH --ntasks-per-node=1
#SBATCH --time=2:00:00    
#SBATCH --nodes=1          
#SBATCH --job-name={self.SUBMIT_PREFIX}_{job}_{index}
#SBATCH --output={self.FOLDER_HOME}/{index}/scripts/AI_{job}_{index}.out
#SBATCH --error={self.FOLDER_HOME}/{index}/scripts/AI_{job}_{index}.err
"""    
        
    else:

        print(f"{self.SYSTEM} not yet setup for submission of jobs! Please adjust submit_head(), which is part of setup_system.py!")
        sys.exit()



"""
Prepares commands to run RosettaRelax methods to refine protein structures.

Functions:
    prepare_RosettaRelax: Sets up commands for RosettaRelax job submission.
"""

import os 
              
def prepare_RosettaRelax(self, 
                         index,  
                         cmd,
                         PreMatchRelax=False # Placeholder for when RosettaMatch gets implemented! Should be removed!
                        ):
    """
    Relaxes protein structure in {index} using RosettaRelax.
    
    Parameters:
    index (str): The index of the protein variant to be relaxed.
    cmd (str): collection of commands to be run, this script wil append its commands to cmd
    
    Optional Parameters:
    PreMatchRelax (bool): True if ESMfold to be run without ligand (prior to RosettaMatch).

    """
    
    filename = f'{self.FOLDER_DESIGN}/{index}'
        
    # Make directories
    os.makedirs(f"{filename}/scripts", exist_ok=True)
    
    # Options for EXPLORE, accelerated script for testing
    if self.EXPLORE:
        repeats = "1"
        ex = ""
    else:
        repeats = "3"
        ex = "-ex1 -ex2"

    PDB_input = self.all_scores_df.at[int(index), "step_input_variant"]
  
    # Create the RosettaRelax.xml file    
    RosettaRelax_xml = f"""
<ROSETTASCRIPTS>

    <SCOREFXNS>
    
        <ScoreFunction name      = "score"                   weights = "beta_nov16" >
            <Reweight scoretype  = "atom_pair_constraint"    weight  = "1" />
            <Reweight scoretype  = "angle_constraint"        weight  = "1" />
            <Reweight scoretype  = "dihedral_constraint"     weight  = "1" />
        </ScoreFunction> 
        
        <ScoreFunction name      = "score_final"             weights = "beta_nov16" >
            <Reweight scoretype  = "atom_pair_constraint"    weight  = "1" />
            <Reweight scoretype  = "angle_constraint"        weight  = "1" />
            <Reweight scoretype  = "dihedral_constraint"     weight  = "1" />
        </ScoreFunction>
        
    </SCOREFXNS>
       
    <MOVERS>
                                  
        <FastRelax  name="mv_relax" disable_design="false" repeats="{repeats}" /> 
"""
    
    # Add constraint if run is not used as PreMatchRelax
    if self.CST_NAME is not None:
        if not PreMatchRelax: RosettaRelax_xml += f"""
            <AddOrRemoveMatchCsts     name="mv_add_cst" 
                                  cst_instruction="add_new" 
                                  cstfile="{self.FOLDER_PARENT}/{self.CST_NAME}.cst" />

"""
        
    RosettaRelax_xml += f"""

        <InterfaceScoreCalculator   name                   = "mv_inter" 
                                    chains                 = "X" 
                                    scorefxn               = "score_final" />
    </MOVERS>
    
    <PROTOCOLS>  

        <Add mover_name="mv_relax" />
"""

    if self.CST_NAME is not None:
        RosettaRelax_xml += f"""                                  
        <Add mover_name="mv_add_cst" /> """
    RosettaRelax_xml += f""" 
        <Add mover_name="mv_inter" />
"""
        
    RosettaRelax_xml += f"""
    </PROTOCOLS>
    
</ROSETTASCRIPTS>
"""
    
    cmd += f"""### RosettaRelax ###
"""
        
    # Write the RosettaRelax.xml to a file
    with open(f'{filename}/scripts/RosettaRelax_{index}.xml', 'w') as f:
        f.writelines(RosettaRelax_xml)   
  
    cmd += f"""
# Run Rosetta Relax
{self.ROSETTA_PATH}/bin/rosetta_scripts.{self.rosetta_ext} \\
    -s                                        {PDB_input}.pdb \\
    -parser:protocol                          {filename}/scripts/RosettaRelax_{index}.xml \\
    -out:file:scorefile                       {filename}/score_RosettaRelax.sc \\
    -nstruct                                  1 \\
    -ignore_zero_occupancy                    false \\
    -corrections::beta_nov16                  true \\
    -run:preserve_header                      true """
    if self.LIGAND not in ['HEM']:
        cmd += f"""\\
    -extra_res_fa                             {self.FOLDER_INPUT}/{self.LIGAND}.params """
    cmd += f"""\\
    -overwrite {ex}

# Cleanup
mv {self.FOLDER_DESIGN}/{index}/{os.path.basename(PDB_input)}_0001.pdb \\
   {self.FOLDER_DESIGN}/{index}/{self.WT}_RosettaRelax_{index}.pdb
sed -i '/        H  /d' {self.FOLDER_DESIGN}/{index}/{self.WT}_RosettaRelax_{index}.pdb
"""
    
    return cmd


"""
The AIzymes script defines the main AIzymes workflow, including setup, initialization, control, and plotting functions.
It manages the primary processes and configurations required to execute AIzymes functionalities.
"""
    
from main_running_003         import *
from main_startup_002         import *
from plotting_002             import *
from helper_002               import *

class AIzymes_MAIN:
    """
    Main class for managing AIzymes workflow, including setup, initialization, control, and plotting functions.

    Functions:
        __init__():     Initializes an instance of the AIzymes_MAIN class.
        setup():        Sets up the AIzymes project environment with specified parameters.
        initialize():   Initializes AIzymes with provided configurations.
        controller():   Controls the AIzymes project based on scoring and normalization parameters.
        plot():         Generates various plots based on AIzymes data.
    """

    def __init__(self, FOLDER_HOME, UNBLOCK_ALL=False, PRINT_VAR=False, PLOT_DATA=False, LOG='info'):
        """
        Initializes AIzymes with given parameters.

        Parameters:
            FOLDER_HOME (str):  Path to the main folder.
            UNBLOCK_ALL (bool): Flag to unblock all processes.
            PRINT_VAR (bool):   Flag to print variables.
            PLOT_DATA (bool):   Flag to plot data.
            LOG (str):          Logging level.
        """
        for key, value in locals().items():
            if key not in ['self']:  
                setattr(self, key, value)

        if os.path.isdir(FOLDER_HOME):
            initialize_controller(self, FOLDER_HOME)
        else:
            print(f"Folder {FOLDER_HOME} missing. Please run setup().")
                  
    def setup(self, FOLDER_HOME, WT, LIGAND, DESIGN,

              # General Job Settings
              MAX_JOBS            = 100, 
              MAX_GPUS            = 0,
              MEMORY              = 128,
              N_PARENT_JOBS       = 3, 
              MAX_DESIGNS         = 10000, 
              KBT_BOLTZMANN       = [0.5, 0.0003],
              MATCH               = None, 
              
              # System Settings
              SUBMIT_PREFIX       = None, 
              SYSTEM              = None,
              RUN_PARALLEL        = False, 
              LOG                 = 'info',   
              REMOVE_TMP_FILES    = False,
              
              # General Design Settings
              PARENT_DES_MED      = ['RosettaDesign','ElectricFields'],
              DESIGN_METHODS      = [[0.7,'RosettaDesign','ElectricFields'],\
                                     [0.3,'ProteinMPNN','ESMfold','RosettaRelax','ElectricFields']],
              EXPLORE             = False,
              RESTRICT_RESIDUES   = None, #2D list defining restricted residues in 1 letter code, e.g.: [[99,'DE'],[103,'H']]
              FOLDER_PARENT       = 'parent',
              FOLDER_PAR_STRUC    = None,
              
              # General Scoring Settings
              SCORING_METHODS     = ['MDMin','ESMfold','RosettaRelax','ElectricFields'], 
              SELECTED_SCORES     = ["total","catalytic","interface","efield"],
              MDMin               = False, 
              WEIGHT_TOTAL        = 1.0,
              WEIGHT_CATALYTIC    = 1.0,
              WEIGHT_INTERFACE    = 1.0,
              WEIGHT_IDENTICAL    = 1.0,
              
              # RosettaDesign Settings
              CST_WEIGHT          = 1.0, 
              CST_NAME            = None,
              
              # ProteinMPNN settings
              ProteinMPNN_BIAS    = 0.5, 
              ProteinMPNN_T       = "0.1", 

              # LigandMPNN settings
              LigandMPNN_BIAS     = 0.5, 
              LigandMPNN_T        = "0.1", 
              
              # SolubleMPNN settings
              SolubleMPNN_BIAS    = 0.5, 
              SolubleMPNN_T       = "0.1", 

              # FieldTools settings
              FIELD_TARGET        = None,
              FIELDS_EXCL_CAT     = True,
              WEIGHT_EFIELD       = 1.0,

              # BioDC settings
              TARGET_REDOX        = 10,
              WEIGHT_REDOX        = 1.0,
              
              # RosettaMatch settings
              FOLDER_MATCH        = None,

              # Identical score settings
              IDENTICAL_DESIGN    = False,

              # Catalytic score settings
              CST_DIST_CUTOFF     = False,
              
              # Established Modules list
              # All Methods that redesign a sequence
              SYS_DESIGN_METHODS  = ["RosettaDesign","ProteinMPNN","LigandMPNN","SolubleMPNN"],
              # All Methods that create a structure
              SYS_STRUCT_METHODS  = ["RosettaDesign","MDMin","ESMfold","RosettaRelax",'AlphaFold3INF'], 
              # All Methods that require GPUs
              SYS_GPU_METHODS     = ["ESMfold",'AlphaFold3INF',"ProteinMPNN","LigandMPNN","SolubleMPNN"],
              ):
        """
        Sets up the AIzymes project environment with specified parameters.

        Parameters:
            FOLDER_HOME (str):          Path to the main folder.
            WT (str):                   Wild type information.
            LIGAND (str):               Ligand data.
            DESIGN (str):               Design specifications.
            MAX_JOBS (int):             Maximum number of jobs to run concurrently.
            MAX_GPUS (int):             Maximum number of GPUs to use.
            MEMORY (int):               Allocated memory in MB.
            N_PARENT_JOBS (int):        Number of parent jobs.
            MAX_DESIGNS (int):          Maximum number of designs.
            KBT_BOLTZMANN (list):       Boltzmann constant values.
            MATCH (str):                Match specifications.
            SUBMIT_PREFIX (str):        Submission prefix.
            SYSTEM (str):               System information.
            RUN_PARALLEL (bool):        Whether to run jobs in parallel.
            LOG (str):                  Logging level.
            REMOVE_TMP_FILES (bool):    Whether to remove temporary files.
            PARENT_DES_MED (list):      List of parent design methods.
            DESIGN_METHODS (list):      List of design methods with proportions.
            EXPLORE (bool):             Whether to explore parameter space.
            RESTRICT_RESIDUES (list):   List defining restricted residues.
            FOLDER_PARENT (str):        Parent folder name.
            FOLDER_PAR_STRUC (str):     Folder for parent structures.
            SCORING_METHODS (list):     List of scoring methods.
            SELECTED_SCORES (list):     List of selected scores.
            MDMin (bool):               Whether to use MD minimization.
            CST_WEIGHT (float):         Weight for constraints.
            CST_NAME (str):             Name for constraint.
            ProteinMPNN_BIAS (float):   Bias for ProteinMPNN.
            ProteinMPNN_T (str):        Temperature for ProteinMPNN.
            LigandMPNN_BIAS (float):    Bias for LigandMPNN.
            LigandMPNN_T (str):         Temperature for LigandMPNN.
            SolubleMPNN_BIAS (float):   Bias for SolubleMPNN.
            SolubleMPNN_T (str):        Temperature for SolubleMPNN.
            FIELD_TARGET (str):         Target for field calculations.
            FIELDS_EXCL_CAT (bool):     Exclude catalytic field.
            TARGET_REDOX (int):         Target redox value.
            FOLDER_MATCH (str):         Folder for match files.
            IDENTICAL_DESIGN (bool):    Whether designs are identical.
            CST_DIST_CUTOFF (bool):     Whether to apply a distance cutoff for constraints.
            SYS_DESIGN_METHODS (list):  List of design methods.
            SYS_STRUCT_METHODS (list):  List of structure creation methods.
            SYS_GPU_METHODS (list):     List of GPU-required methods.
        """
        for key, value in locals().items():
            if key not in ['self']:  
                setattr(self, key, value)
        
        aizymes_setup(self)

    def controller(self):
        """
        Main script to run AIzymes.
        """
        for key, value in locals().items():
            if key not in ['self']:
                setattr(self, key, value)
        start_controller(self)

    def submit_controller(self):
        """
        Creates submition script to start the controller.
        """
        submit_controller_parallel(self)
        
    def plot(self, 
             SCORES_V_INDEX=False, 
             STATISTICS=False,
             SCORES_V_GEN=False,
             SCORES_HIST=False,
             PRINT_VALS=False, 
             RESOURCE_LOG=False,
             NORM={},
             HIGHSCORE={},
             NEGBEST={},
             PLOT_TREE=False,
             landscape_plot=False,
             PLOT_SIZE=3,
             TREE_SCORE="final_score"):
        """
        Generates plots based on AIzymes data, including main, tree, and landscape plots.

        Parameters:
            SCORES_V_INDEX (bool):  Flag to generate plots scores vs index.
            STATISTICS (bool):      Flag to generate statistical plots.
            SCORES_V_GEN (bool):    Flag to generate plots scores vs generation.
            SCORES_HIST (bool):     Flag to generate histogram plots.
            PRINT_VALS (bool):      Flag to print values on the plots.
            RESOURCE_LOG (bool):    Flag to log resource usage.
            NORM (dict):            Normalization parameters.
            HIGHSCORE (dict):       High score values.
            NEGBEST (dict):         Negative best score values.
            PLOT_TREE (bool):       Flag to generate tree plots.
            landscape_plot (bool):  Flag to generate landscape plots.
            PLOT_SIZE (int):        Size parameter for plots.
            TREE_SCORE (str):       Score metric used for tree plots.
        """
        for key, value in locals().items():
            if key not in ['self']:
                setattr(self, key, value)
        make_plots(self)
        return

    def best_structures(self,
                        SEQ_PER_ACTIVE_SITE = None, 
                        ACTIVE_SITE         = None,
                        N_HITS              = 50,
                        PLOT_SIZE           = 3):
        """
        Retrieves the best structures based on active site criteria and scoring.

        Parameters:
            SEQ_PER_ACTIVE_SITE (int):  Number of sequences per active site.
            ACTIVE_SITE (str):          Comma-seperated list of active site residues.
            N_HITS (int):               Number of top hits to retrieve.
            PLOT_SIZE (int):            Plot size for visual representation.
        """
        for key, value in locals().items():
            if key not in ['self']:
                setattr(self, key, value)
        get_best_structures(self)


import sys
import os
import subprocess
import numpy as np  
import pickle
import datetime
import mdtraj as md
import parmed as pmd

def ResNr_from_ResName(ResName, Names):
    for Name_i in range(0,len(Names),1):
        if ResName == Names[Name_i][1]:
            return Names[Name_i][2]

def Calc_Electrostatic_Potential_Energy(_ABC,_XYZ,_Charge,_Target_Charge):
    # Electrostatic Potential energy at _xyz from Atom _abc with _Charge and Target_Charges U = k*q*Q/r
    _Couloumb_Const = 1.00 #(A.U.)
    _Vector = _ABC-_XYZ        
    _Length = np.linalg.norm(_Vector)
    _Length_Bohr = _Length/0.529177249           # Bohr
    _Energy = (_Couloumb_Const*float(_Charge)*float(_Target_Charge)/(_Length_Bohr))   #### WRONG!!!!!!!!!!
    ### 1 a.u. = 2625.5 kJ/mol
    _Energy=_Energy*2625.5    
    return(_Energy)

def Field_of_ABC_at_XYZ_projected_onto_VEC(_ABC,_XYZ,_VEC,_Charge,arg_energy_out,Target_Charges):
    
    # Field at _xyz from Atom _abc with _Charge E = k*Q/r/r
    _Couloumb_Const = 1.00 #(A.U.)
    _Vector = _ABC-_XYZ     
    _Length = np.linalg.norm(_Vector)
    _Normalized_Vector = (-1.0)*_Vector/_Length
    _Length_Bohr = _Length/0.529177249           # Bohr
    _Field = (_Couloumb_Const*float(_Charge)/_Length_Bohr**2)*_Normalized_Vector
    _Field = np.dot(_Field,_VEC)
    ### 1 a.u. = 5.14x10^9 V/cm
    ### 1 a.u. = 5.14x10^3 MV/cm
    _Field=_Field*5.14*1000 / 100

    _Energy = 0.0               
    return _Field,_Energy

def Field_of_ABC_at_XYZ(_ABC,_XYZ,_Charge,arg_energy_out,Target_Charges,arg_vector_out):
    
    # Field at _xyz from Atom _abc with _Charge E = k*Q/r/r 
    _Couloumb_Const = 1.00 #(A.U.)
    _Vector = _ABC-_XYZ     
    _Length = np.linalg.norm(_Vector)
    _Length_Bohr = _Length/0.529177249           # Bohr
    _Field = (-1)*(_Couloumb_Const*float(_Charge)/(_Length_Bohr**2))
    ### 1 a.u. = 5.14x10^9 V/cm
    ### 1 a.u. = 5.14x10^3 MV/cm
    _Field=_Field*5.14*1000 / 100
    
    _Energy = 0.0
    _Field_Vectors = []
    _Field_Vector = "x"
    if arg_energy_out != "False":
        _Energy = Calc_Electrostatic_Potential_Energy(_ABC,_XYZ,_Charge,Target_Charges[0])
    if arg_vector_out != "False": 
        _Normalized_Vector = _Vector/_Length
        _Field_Vector = (-1.0)*(_Couloumb_Const*float(_Charge)/_Length_Bohr**2)*_Normalized_Vector
        #print("Field magnitude:",(-1.0)*(_Couloumb_Const*float(_Charge)/_Length_Bohr**2))
        #print("Field direction:",_Normalized_Vector,"length of vector",np.linalg.norm(_Normalized_Vector))
        #print("Field vector:",_Field_Vector,"length of vector",np.linalg.norm(_Field_Vector))
        
    return _Field,_Energy,_Field_Vector

def fkt_Update_Charges(QM_Charges,QM_Dict,Frame_i,Names,Charges):
    stop="X"
    for QM_Charge_i in range(0,len(QM_Charges),1):
        if len(QM_Charges[QM_Charge_i])==2:
            if QM_Charges[QM_Charge_i][1]==str(Frame_i+1):
                start = QM_Charge_i+1
            if QM_Charges[QM_Charge_i][1]==str(Frame_i+2):
                stop = QM_Charge_i
                QM_Charges = QM_Charges[start:stop]
                break
    if stop == "X":
        QM_Charges = QM_Charges[start:]   
    
    for QM_Atom_i in range(0,len(QM_Dict),1):
        for Atom_i in range(0,len(Names),1):
            Name_tmp = Names[Atom_i][0]+"_"+Names[Atom_i][1]+"_"+Names[Atom_i][2]
            if Name_tmp == QM_Dict[QM_Atom_i]:      #QM_Atom_i index in QM_Charges to be changed from 
                Charges[Atom_i]=QM_Charges[QM_Atom_i][2]
                break
    return Charges

def fkt_Load_QM_Charges(arg_qm_charges,arg_qm_dict):
    with open(arg_qm_charges) as f:
        QM_Charges = [i.split() for i in f.readlines()]
    with open(arg_qm_dict) as f:
        QM_Dict = [i.split("\n")[0] for i in f.readlines()]
    return QM_Charges,QM_Dict

def decompose_fields(Absolut_Fields,Absolut_Field,Names,Atom_Index,arg_solvent):
    #### Total Field
    Absolut_Fields["Total"]+=Absolut_Field
    #### Solvent Field
    if Names[Atom_Index][1] in arg_solvent:
        Absolut_Fields["Solvent"]+=Absolut_Field
        Absolut_Fields[Names[Atom_Index][1]]+=Absolut_Field
    else:
        Absolut_Fields["Protein"]+=Absolut_Field
        Absolut_Fields[Names[Atom_Index][1]+"_"+Names[Atom_Index][2]]+=Absolut_Field
    return Absolut_Fields

def fkt_calc_fields(FIELDS,FIELD,XYZ,VEC,Frame,Charges,Names,Field_Components,arg_solvent,Self_i,arg_exclude_atoms, \
                    arg_energy_out,Target_Index,ENERGIES,arg_vector_out,VECTORS):

### VARIABLES STRUCTURE
    Absolut_Fields = {}
    for Field_Component in Field_Components:
        Absolut_Fields[Field_Component] = 0.0
    Absolut_Energies = {}
    for Field_Component in Field_Components:
        Absolut_Energies[Field_Component] = 0.0

    Target_Charges = Target_Index[FIELD]
    Target_Charges = [Charges[i] for i in Target_Charges]
    for Atom_Index in range(0,len(Frame),1):   
#### Exclude Self-residue!!! 
        if arg_exclude_atoms == "False":
            if not Self_i.isdigit():
                Self_i = ResNr_from_ResName(Self_i,Names)
            if Names[Atom_Index][2] == Self_i: continue        
        else:
            if Atom_Index in arg_exclude_atoms[FIELD]: continue
                
                
        #### Field_of_ABC_at_XYZ(_ABC,_XYZ,_Charge,arg_energy_out,arg_vector_out,Target_Charges):
        if len(FIELD.split("_")) == 1:   #POINT CALCULATION
            Absolut_Field, Absolut_Energy, Vectors  = Field_of_ABC_at_XYZ(Frame[Atom_Index],XYZ,Charges[Atom_Index], \
                                                                   arg_energy_out, Target_Charges, arg_vector_out)
            
        ### Field_of_ABC_at_XYZ_projected_onto_VEC(_ABC,_XYZ,_VEC,_Charge,arg_energy_out,Target_Charges)
        if len(FIELD.split("_")) == 2:   #VECTOR CALCULATION
            Absolut_Field, Absolut_Energy = Field_of_ABC_at_XYZ_projected_onto_VEC(Frame[Atom_Index],XYZ,VEC,\
                                                                                   Charges[Atom_Index],arg_energy_out,\
                                                                                   Target_Charges)
      
        Absolut_Energies      = decompose_fields(Absolut_Energies,Absolut_Energy,Names,Atom_Index,arg_solvent)
        Absolut_Fields        = decompose_fields(Absolut_Fields,Absolut_Field,Names,Atom_Index,arg_solvent)
        
    for Field_Component in Field_Components:
        FIELDS[FIELD][Field_Component].append(Absolut_Fields[Field_Component])
        if arg_energy_out != "False":
            ENERGIES[FIELD][Field_Component].append(Absolut_Energies[Field_Component])
    if len(FIELD.split("_")) == 1: 
        if arg_vector_out != "False":
            VECTORS[FIELD].append(Vectors)
    return FIELDS,ENERGIES,VECTORS

def fkt_Field_Components(FIELDS,Names,arg_solvent):
    
    Field_Components = ["Total","Protein","Solvent"]+arg_solvent
    
    for Name in Names:
        if Name[1] not in arg_solvent:
            if Name[1]+"_"+Name[2] not in Field_Components: Field_Components += [Name[1]+"_"+Name[2]] 

    for Field in FIELDS:
        for Field_Component in Field_Components:
            FIELDS[Field][Field_Component]=[]
    
    return FIELDS,Field_Components

def fkt_get_Target_VEC(Target_XYZ):
    
    Target_VEC = {}

    for Target in Target_XYZ:
        if len(Target_XYZ[Target])==2:
            Target_VEC[Target]={}
            XYZ_START = Target_XYZ[Target][0]
            XYZ_STOP  = Target_XYZ[Target][1]
            Target_VEC[Target]["Center"] = (XYZ_START+XYZ_STOP)/2 
            Target_VEC[Target]["Vector"] = (XYZ_STOP-XYZ_START)/np.linalg.norm(XYZ_STOP-XYZ_START)
            
    return Target_VEC

def fkt_Target_Index(Names,arg_target,FIELDS):
    
    Target_Index={}
    
    with open(arg_target) as f:
        Targets = [i.split() for i in f.readlines()]
        
    for Target in range(len(Targets)-1,-1,-1): #Delete empty lines
        if len(Targets[Target]) == 0:
            Targets = Targets[:Target]+Targets[Target+1:]
    
    for Target_i in range(0,len(Targets),1):
        Target_Name = "_".join(Targets[Target_i])
        FIELDS[Target_Name]={}
        for Atom_i in range(0,len(Targets[Target_i]),1): #If this is a vector, it contains two Atoms!
            Atom_ResidueNr = Targets[Target_i][Atom_i].split("@")[0][1:]
            Atom_Name      = Targets[Target_i][Atom_i].split("@")[1]
            #print Targets[Target_i][Atom_i], Atom_ResidueNr, Atom_Name
            
            if not Atom_ResidueNr.isdigit(): Atom_ResidueNr = ResNr_from_ResName(Atom_ResidueNr,Names)
    
            for Name_i in range(0,len(Names),1):
                if Atom_ResidueNr == Names[Name_i][2]:      #Check if Residue Number matches
                    if Atom_Name == Names[Name_i][0]:       #Check if Atom Name matches
                        Targets[Target_i][Atom_i] = Name_i  #Update Number in Target
                        break
        Target_Index[Target_Name] = Targets[Target_i]
    return Target_Index,FIELDS #Passed to Target_Index in main!

def fkt_Exlude_Index(Names,arg_exclude_atoms,arg_target):
    
    Exclude_Index={}
    
    with open(arg_exclude_atoms) as f:
        Excludes = [i.split() for i in f.readlines()]
    with open(arg_target) as f:
        Targets = [i.split() for i in f.readlines()]

    for Excludes_i in range(0,len(Excludes),1):
        Exclude_Index["_".join(Targets[Excludes_i])]=Excludes[Excludes_i]
        for Atom_i in range(0,len(Excludes[Excludes_i]),1):    #If this is a vector, it contains two Atoms!
            Atom_ResidueNr = Excludes[Excludes_i][Atom_i].split("@")[0][1:]
            if not Atom_ResidueNr.isdigit(): Atom_ResidueNr = ResNr_from_ResName(Atom_ResidueNr,Names)
            Atom_Name      = Excludes[Excludes_i][Atom_i].split("@")[1]
            for Name_i in range(0,len(Names),1):
                if Atom_ResidueNr == Names[Name_i][2]:      #Check if Residue Number matches
                    if Atom_Name == Names[Name_i][0]:       #Check if Atom Name matches
                        Exclude_Index["_".join(Targets[Excludes_i])][Atom_i] = Name_i  #Update Number in Target
                        break
    arg_exclude_atoms =  Exclude_Index              
    return arg_exclude_atoms #Passed to Target_Index in main!

def fkt_get_Target_XYZ(Frame,Target_Index):
#Target_Index is a libary
#Frame is
#Make empty library
    Target_XYZ = {}

    for Target in Target_Index:
        Target_XYZ[Target]=[]
        for Atom_i in range(0,len(Target_Index[Target]),1): #If this is a vector, it contains two Atoms!
            Target_XYZ[Target]=Target_XYZ[Target]+[Frame[Target_Index[Target][Atom_i]]]
                
    return Target_XYZ #Passed to Target_Index in main!

def fkt_Load_Trajectory(arg_nc, arg_parm, arg_TIP4P):
    
    # Load topology using ParmEd
    parm = pmd.load_file(arg_parm)
    
    # Extract charges from ParmEd topology
    Charges = np.array([atom.charge for atom in parm.atoms])
    
    # Load trajectory using MDTraj
    traj = md.load(arg_nc, top=arg_parm)
    Trajectory = traj.xyz  # Coordinates in nanometers
    
    # Extract atomic and residue information
    Names = [[str(atom.name), str(atom.residue.name), str(atom.residue.index + 1)] for atom in traj.topology.atoms]

    if arg_TIP4P == "True":
        TIP4P_indices = [
            i for i, name in enumerate(atom.name for atom in traj.topology.atoms)
            if name == "EPW" and atom.residue.name == "WAT"
        ]
        for index in TIP4P_indices:
            Charges[index - 3] = np.inf  # Ignore charge of linker atoms
            for frame in Trajectory:
                frame[index - 3] = [np.inf, np.inf, np.inf]

        Trajectory = np.array([
            frame[np.all(frame != [np.inf, np.inf, np.inf], axis=1)]
            for frame in Trajectory
        ])
        Charges = Charges[np.isfinite(Charges)]

    return Trajectory, Charges, np.array(Names)

def make_pdb(arg_nc, arg_parm, arg_pdb, arg_use_qmcharges, arg_TIP4P, arg_qm_mask, arg_out):
    traj = md.load(arg_nc, top=arg_parm)
    
    if arg_TIP4P == "True":
        TIP4P_indices = [
            i for i, atom in enumerate(traj.topology.atoms)
            if atom.name == "EPW" and atom.residue.name == "WAT"
        ]
        # Remove extra points from both topology and trajectory
        traj = traj.atom_slice([i for i in range(traj.n_atoms) if i not in TIP4P_indices])
    
    traj.save_pdb(arg_pdb)
        
def load_qm_mask(arg_qm_mask):
    with open(arg_qm_mask) as f:
        arg_qm_mask = f.read()[:-1]
    return arg_qm_mask

def usage():
    print("                                                                                                              ")
    print("|------------------------------------------------------------------------------------------------------------|")
    print("|-------------                                FieldTools Usage:                                 -------------|")
    print("|------------------------------------------------------------------------------------------------------------|")
    print("|-nc             : Specify trajectory file                                                                   |")
    print("|-parm           : Specify parameter file                                                                    |")
    print("|-out            : Specify output file                                                                       |")
    print("|-energy_out     : Optional: Calculate Energies from Fields                                                  |")
    print("|-target         : Specify target file                                                                       |")
    print("|-solvent        : Optional: Comma seperated list of non-protein residues              [default WAT,Na+,Cl- ]|")
    print("|-exclude_atoms  : Optional: Specify what atoms will be excluded from Field calculation                      |")
    print("|                : If not set, Field calculation will exclude the full residue of the first atom defined     |")
    print("|                : for each target.                                                                          |")
    print("|-TIP4P          : Optional: Recognize TIP4P waters                                    [default False       ]|")
    print("|-verbose        : Optional: Display additional information                            [default False       ]|")
    print("|-use_qm_charges : Optional: Specify if qmcharges should be read                       [default False       ]|")
    print("|-vector_out     : Optional: Save Field Vectors to specified file                      [default False       ]|")
    print("|------------------------------------------------------------------------------------------------------------|")
    print("|-------------        If -use_qm_charges = true, the following options have to be set:          -------------|")
    print("|------------------------------------------------------------------------------------------------------------|")
    print("|-qm_mask        : Specify qm region (amber selection mask)                                                  |")
    print("|-qm_charges     : Specify qmcharges file (qm partial charges of each frame, made with QMChargesTools)       |")
    print("|-qm_dict        : Specify qmdict file (containing qmatom names in .pdb and gaussian)                        |")
    print("|------------------------------------------------------------------------------------------------------------|")
    print("|-------------                              How to define -target:                              -------------|")
    print("|------------------------------------------------------------------------------------------------------------|")
    print("|-target should point to a file, each line in that file defines target to calculate the Field                |")
    print("|Targets are selected using the amber selection syntax                                                       |")
    print("|To calculate the magnitude of the field at a specific atom:     :47@OG                                      |")
    print("|To calculate the magnitude of the field projected along a bond: :47@OG :47@OH                               |")
    print("|------------------------------------------------------------------------------------------------------------|")
    print("|-------------                                      INFO                                        -------------|")
    print("|------------------------------------------------------------------------------------------------------------|")
    print("|Script recognizes if total field at one point or projection onto a bond is to be calculated                 |")
    print("|based on the input in -target                                                                               |")
    print("|-------------                                                                                               |")
    print("|Script excludes all Field contributions of the Residue at which the field is calculated.                    |")
    print("|For Vector Projections, this is only the first residue number!                                              |")
    print("|-------------                                                                                               |")
    print("|Solvent = All atoms as defined in -solvent,                                                                 |")
    print("|          for these residues contributions will also be grouped based on the residue name                   |")
    print("|Protein = All atoms that are not part of -solvent,                                                          |")
    print("|          for the protein, per-reside fields will be calculated                                             |")
    print("|-------------                                                                                               |")
    print("|Linker Atom charge is ignored!                                                                              |")
    print("|------------------------------------------------------------------------------------------------------------|") 
    
def inputParser(arg):
    
    arg_nc            = False                                                                         ### Add new flag here
    arg_parm         = False
    arg_out           = False
    arg_energy_out    = "False"
    arg_target        = False
    arg_solvent       = "WAT,Na+,Cl-"
    arg_exclude_atoms = "False" 
    arg_TIP4P         = "False"
    arg_verbose       = "False"
    arg_use_qmcharges = "False"
    arg_vector_out    = "False"
    arg_qm_mask       = False
    arg_qm_charges    = False
    arg_qm_dict       = False  
    
    if(len(arg) == 1) or arg[1] in ["-help","--help","-h"]:
        usage()
        quit()
    
    for index in range(1,len(arg)-1,2):
        param = arg[index]
        if param=="-nc"               : arg_nc            = str(arg[index+1])         ### Add new flag here
        elif param=="-parm"           : arg_parm          = str(arg[index+1])
        elif param=="-out"            : arg_out           = str(arg[index+1])
        elif param=="-energy_out"     : arg_energy_out    = str(arg[index+1])
        elif param=="-target"         : arg_target        = str(arg[index+1])
        elif param=="-solvent"        : arg_solvent       = str(arg[index+1])
        elif param=="-exclude_atoms"  : arg_exclude_atoms = str(arg[index+1])
        elif param=="-qm_dict"        : arg_qm_dict       = str(arg[index+1])
        elif param=="-TIP4P"          : arg_TIP4P         = str(arg[index+1])
        elif param=="-verbose"        : arg_verbose       = str(arg[index+1])
        elif param=="-use_qm_charges" : arg_use_qmcharges = str(arg[index+1])
        elif param=="-vector_out"     : arg_vector_out    = str(arg[index+1])
        elif param=="-qm_mask"        : arg_qm_mask       = str(arg[index+1])
        elif param=="-qm_charges"     : arg_qm_charges    = str(arg[index+1])
        elif param=="-qm_dict"        : arg_qm_dict       = str(arg[index+1])
        else:
            print ("Error! "+param+" not a valid option.")
            usage()
            quit()  
                                                                                                      ### Add new flag here
    if arg_nc        == False: print ("ERROR. Please specify the name of the trajectory file with -nc")  
    if arg_parm     == False: print ("ERROR. Please specify the name of the parameter file with -parm")
    if arg_out       == False: print ("ERROR. Please specify the name of the output file with -out")
    if arg_target    == False: print ("ERROR. Please specify the name of the target file with -target.                                       \n -> See examples with --help")
    if arg_use_qmcharges == "True":
        if arg_qm_mask    == False: print ("ERROR. Please specify qm mask with -qm_mask")
        if arg_qm_charges == False: print ("ERROR. Please specify charge of the qm region with -qm_charge")
        if arg_qm_dict    == False: print ("ERROR. Please specify qm_dict with -qm_dict")
        
    if False in [arg_nc,arg_parm,arg_out,arg_target]:                                                ### Add new flag here
        usage()
        quit()
    if arg_use_qmcharges == "True":
        if False in [arg_qm_mask,arg_qm_charges,arg_qm_dict]:                                         ### Add new flag here
            usage()
            quit()
    
    print ("-nc              Trajectory file      : ",arg_nc)                                         ### Add new flag here
    print ("-parm            Parameter file       : ",arg_parm)
    print ("-out             Output file          : ",arg_out)
    print ("-target          Field target         : ",arg_target)
    print ("-arg_solvent     Non-protein residues : ",arg_solvent)
    if arg_exclude_atoms != "False":
        print ("-exclude_atoms   Atoms excluded       : ",arg_exclude_atoms)
    else:
        print ("-exclude_atoms   Atoms excluded       : Full residue defined in the first Atom for each Target.")
    if arg_TIP4P == "True":
        print ("-TIP4P           Recognize TIP4P wat  : ",arg_TIP4P)
    if arg_energy_out != "False":
        print ("-energy_out      Energy from Fields   : ",arg_energy_out)
    if arg_verbose == "True":
        print ("-arg_verbose     Verbose information  : ",arg_verbose)
    if arg_use_qmcharges == "True":
        print ("-use_qm_charges  QM_theory            :  True")
        print ("-qm_mask         QM region            : ",arg_qm_mask)
        print ("-qm_charges      QM charges           : ",arg_qm_charges)
        print ("-qm_dict         QM dict              : ",arg_qm_dict) 
    if arg_vector_out != "False":
        print ("-vector_out      Vector output file   : ",arg_vector_out)
    return arg_nc,arg_parm,arg_out,arg_energy_out,arg_target,arg_solvent,arg_exclude_atoms,arg_TIP4P,arg_verbose,      \
           arg_use_qmcharges,arg_qm_mask,arg_qm_charges,arg_qm_dict,arg_vector_out                    ### Add new flag here

def python_main(args):                                                                           
    arg_nc,arg_parm,arg_out,arg_energy_out,arg_target,arg_solvent,arg_exclude_atoms,arg_TIP4P,arg_verbose,       \
    arg_use_qmcharges,arg_qm_mask,arg_qm_charges,arg_qm_dict,arg_vector_out=inputParser(args)     ### Add new flag here
    arg_solvent = arg_solvent.split(",")

    print ("\nField calculation RUNNING : ",datetime.datetime.now()) 

    #Load qm_mask.in
    if arg_use_qmcharges == "True":
        arg_qm_mask = load_qm_mask(arg_qm_mask)

    ### Empty dictionary storing all Fields
    FIELDS = {}
    ENERGIES = {}
    VECTORS = {}
    
######################## not necessary anymore! Keept in case this might become interesting again
    ### Make pdb of system
    #arg_pdb = ".".join(arg_out.split(".")[:-1])+"_full.pdb"
    #make_pdb(arg_nc,arg_param,arg_pdb,arg_use_qmcharges,arg_TIP4P,arg_qm_mask,arg_out)
######################## not necessary anymore!

    ### Load QM charges
    if arg_use_qmcharges == "True":
        QM_Charges,QM_Dict = fkt_Load_QM_Charges(arg_qm_charges,arg_qm_dict) 
        if arg_verbose=="True": print ("QM_Dict :", QM_Dict,"\n" )     
        if arg_verbose=="True": print ("QM_Charges :", QM_Charges[:50],"\n")

    ### Load Trajectory
    Trajectory, Charges, Names = fkt_Load_Trajectory(arg_nc,arg_parm,arg_TIP4P)
    if arg_verbose=="True": print ("Trajectory :", Trajectory[0])
    if arg_verbose=="True": print ("Charges :", Charges)
    if arg_verbose=="True": print ("Names :", Names)
    if len(Trajectory[0]) != len(Charges) or len(Trajectory[0]) != len(Names):
        print ("Error! Trajectory, Charges, and Atom Name Lists have different length")
        print ("Trajectory :", len(Trajectory[0]))
        print ("Charges :", len(Charges))
        print ("Names :", len(Names),"\n","\n")
        usage()
        quit()
    
    if arg_verbose=="True": print ("Topology information (index,Coords,Charges,Names):")
    for i in range(0,len(Names),1):
        if arg_verbose=="True": print (i, Trajectory[0][i], Charges[i], Names[i] )

    ### Load Index of Target Atoms for Field calculation
    Target_Index,FIELDS = fkt_Target_Index(Names,arg_target,FIELDS)
    if arg_verbose=="True": print ("Target_Index:", Target_Index,"\n")
    if arg_verbose=="True": print ("Field_Dict:", FIELDS,"\n")

    ### Load Index of Excluded Atoms
    if arg_exclude_atoms != "False":
        arg_exclude_atoms = fkt_Exlude_Index(Names,arg_exclude_atoms,arg_target)
        if arg_verbose=="True": print ("Excluded_Atoms:", arg_exclude_atoms,"\n")
            
    ### Define Field Components (Total,Protein,Solvent,Each residue,Each solvent type) to calculate
    FIELDS,Field_Components = fkt_Field_Components(FIELDS,Names,arg_solvent)
    if arg_verbose=="True": print ("Field_Dict:", FIELDS,"\n")
    if arg_energy_out != "False":
        for FIELD in FIELDS.keys():
            ENERGIES[FIELD] = {}
            for Field_Component in FIELDS[FIELD].keys():
                ENERGIES[FIELD][Field_Component] = []
        if arg_verbose=="True": print ("Energy_Dict:", ENERGIES,"\n") 
    if arg_vector_out != "False":
        for FIELD in FIELDS.keys():
            VECTORS[FIELD] = []
           
        if arg_verbose=="True": print ("Vector_Dict:", VECTORS,"\n") 

    ### Loop through Trajectory to calculate Fields
    for Frame_i in range(0,len(Trajectory),1):
        
    ### Get XYZ coordinates of the Target Atoms
        Target_XYZ=fkt_get_Target_XYZ(Trajectory[Frame_i],Target_Index)
        if arg_verbose=="True": print ("Target_XYZ:",Target_XYZ,"\n")
        Target_VEC=fkt_get_Target_VEC(Target_XYZ)
        if arg_verbose=="True": print ("Target_VEC:",Target_VEC,"\n")    
    ### Update QM_Charges
        if arg_use_qmcharges == "True":
            Charges = fkt_Update_Charges(QM_Charges,QM_Dict,Frame_i,Names,Charges)  
            if arg_verbose=="True": print ("Charges :", Charges,"\n")    
            
    ### Loop through FIELDS to calculate each Target
        for FIELD in FIELDS:
            if len(FIELD.split("_")) == 1:   #POINT CALCULATION
                XYZ = Target_XYZ[FIELD][0]
                VEC = ""
            if len(FIELD.split("_")) == 2:   #VECTOR CALCULATION
                XYZ = Target_VEC[FIELD]['Center']
                VEC = Target_VEC[FIELD]['Vector']
            Self_i = FIELD.split("@")[0][1:]
            
    ### Calculate Fields - This is where the magic happens!
            FIELDS,ENERGIES,VECTORS = fkt_calc_fields(FIELDS,FIELD,XYZ,VEC,Trajectory[Frame_i],Charges,Names, \
                                              Field_Components,arg_solvent,Self_i,arg_exclude_atoms,arg_energy_out, \
                                              Target_Index,ENERGIES, arg_vector_out, VECTORS)
            if arg_verbose=="True": 
                for Field_Component in FIELDS[FIELD]:
                    print (FIELD,Field_Component,FIELDS[FIELD][Field_Component]) 
                if arg_energy_out != "False":    
                    for Field_Component in ENERGIES[FIELD]:
                        print (FIELD,Field_Component,ENERGIES[FIELD][Field_Component])    
    
    pickle.dump(FIELDS,open(arg_out,"wb"))   
    if arg_energy_out != "False":
        pickle.dump(ENERGIES,open(arg_energy_out,"wb"))
    if arg_vector_out != "False":
        pickle.dump(VECTORS,open(arg_vector_out,"wb"))
    print ("\nField calculation DONE : ",datetime.datetime.now() )      
    
def main():
    ########################
    ### THE FINAL SCRIPT ###
    ########################
    ### Read in arguments                                                                             
    arg_nc,arg_param,arg_out,arg_energy_out,arg_target,arg_solvent,arg_exclude_atoms,arg_TIP4P,arg_verbose,       \
    arg_use_qmcharges,arg_qm_mask,arg_qm_charges,arg_qm_dict,arg_vector_out=inputParser(sys.argv)     ### Add new flag here
    arg_solvent = arg_solvent.split(",")

    print ("\nField calculation RUNNING : ",datetime.datetime.now()) 

    #Load qm_mask.in
    if arg_use_qmcharges == "True":
        arg_qm_mask = load_qm_mask(arg_qm_mask)

    ### Empty dictionary storing all Fields
    FIELDS = {}
    ENERGIES = {}
    VECTORS = {}
    
######################## not necessary anymore! Keept in case this might become interesting again
    ### Make pdb of system
    #arg_pdb = ".".join(arg_out.split(".")[:-1])+"_full.pdb"
    #make_pdb(arg_nc,arg_param,arg_pdb,arg_use_qmcharges,arg_TIP4P,arg_qm_mask,arg_out)
######################## not necessary anymore!

    ### Load QM charges
    if arg_use_qmcharges == "True":
        QM_Charges,QM_Dict = fkt_Load_QM_Charges(arg_qm_charges,arg_qm_dict) 
        if arg_verbose=="True": print ("QM_Dict :", QM_Dict,"\n" )     
        if arg_verbose=="True": print ("QM_Charges :", QM_Charges[:50],"\n")

    ### Load Trajectory
    Trajectory, Charges, Names = fkt_Load_Trajectory(arg_nc,arg_param,arg_TIP4P)
    if arg_verbose=="True": print ("Trajectory :", Trajectory[0])
    if arg_verbose=="True": print ("Charges :", Charges)
    if arg_verbose=="True": print ("Names :", Names)
    if len(Trajectory[0]) != len(Charges) or len(Trajectory[0]) != len(Names):
        print ("Error! Trajectory, Charges, and Atom Name Lists have different length")
        print ("Trajectory :", len(Trajectory[0]))
        print ("Charges :", len(Charges))
        print ("Names :", len(Names),"\n","\n")
        usage()
        quit()
    
    if arg_verbose=="True": print ("Topology information (index,Coords,Charges,Names):")
    for i in range(0,len(Names),1):
        if arg_verbose=="True": print (i, Trajectory[0][i], Charges[i], Names[i] )

    ### Load Index of Target Atoms for Field calculation
    Target_Index,FIELDS = fkt_Target_Index(Names,arg_target,FIELDS)
    if arg_verbose=="True": print ("Target_Index:", Target_Index,"\n")
    if arg_verbose=="True": print ("Field_Dict:", FIELDS,"\n")

    ### Load Index of Excluded Atoms
    if arg_exclude_atoms != "False":
        arg_exclude_atoms = fkt_Exlude_Index(Names,arg_exclude_atoms,arg_target)
        if arg_verbose=="True": print ("Excluded_Atoms:", arg_exclude_atoms,"\n")
            
    ### Define Field Components (Total,Protein,Solvent,Each residue,Each solvent type) to calculate
    FIELDS,Field_Components = fkt_Field_Components(FIELDS,Names,arg_solvent)
    if arg_verbose=="True": print ("Field_Dict:", FIELDS,"\n")
    if arg_energy_out != "False":
        for FIELD in FIELDS.keys():
            ENERGIES[FIELD] = {}
            for Field_Component in FIELDS[FIELD].keys():
                ENERGIES[FIELD][Field_Component] = []
        if arg_verbose=="True": print ("Energy_Dict:", ENERGIES,"\n") 
    if arg_vector_out != "False":
        for FIELD in FIELDS.keys():
            VECTORS[FIELD] = []
           
        if arg_verbose=="True": print ("Vector_Dict:", VECTORS,"\n") 

    ### Loop through Trajectory to calculate Fields
    for Frame_i in range(0,len(Trajectory),1):
        
    ### Get XYZ coordinates of the Target Atoms
        Target_XYZ=fkt_get_Target_XYZ(Trajectory[Frame_i],Target_Index)
        if arg_verbose=="True": print ("Target_XYZ:",Target_XYZ,"\n")
        Target_VEC=fkt_get_Target_VEC(Target_XYZ)
        if arg_verbose=="True": print ("Target_VEC:",Target_VEC,"\n")    
    ### Update QM_Charges
        if arg_use_qmcharges == "True":
            Charges = fkt_Update_Charges(QM_Charges,QM_Dict,Frame_i,Names,Charges)  
            if arg_verbose=="True": print ("Charges :", Charges,"\n")    
            
    ### Loop through FIELDS to calculate each Target
        for FIELD in FIELDS:
            if len(FIELD.split("_")) == 1:   #POINT CALCULATION
                XYZ = Target_XYZ[FIELD][0]
                VEC = ""
            if len(FIELD.split("_")) == 2:   #VECTOR CALCULATION
                XYZ = Target_VEC[FIELD]['Center']
                VEC = Target_VEC[FIELD]['Vector']
            Self_i = FIELD.split("@")[0][1:]
            
    ### Calculate Fields - This is where the magic happens!
            FIELDS,ENERGIES,VECTORS = fkt_calc_fields(FIELDS,FIELD,XYZ,VEC,Trajectory[Frame_i],Charges,Names, \
                                              Field_Components,arg_solvent,Self_i,arg_exclude_atoms,arg_energy_out, \
                                              Target_Index,ENERGIES, arg_vector_out, VECTORS)
            if arg_verbose=="True": 
                for Field_Component in FIELDS[FIELD]:
                    print (FIELD,Field_Component,FIELDS[FIELD][Field_Component]) 
                if arg_energy_out != "False":    
                    for Field_Component in ENERGIES[FIELD]:
                        print (FIELD,Field_Component,ENERGIES[FIELD][Field_Component])    
    
    pickle.dump(FIELDS,open(arg_out,"wb"))   
    if arg_energy_out != "False":
        pickle.dump(ENERGIES,open(arg_energy_out,"wb"))
    if arg_vector_out != "False":
        pickle.dump(VECTORS,open(arg_vector_out,"wb"))
    print ("\nField calculation DONE : ",datetime.datetime.now() )              
            
if __name__ == "__main__":
    main()


def make_main_scripts(self):

    # ------------------------------------------------------------------------------------------------------------------------------
    # Create the cif_to_pdb.py script
    # ------------------------------------------------------------------------------------------------------------------------------
    with open(f'{self.FOLDER_PARENT}/cif_to_pdb.py', 'w') as f:
        f.write("""
import gemmi
import argparse

def main(args):

    cif_file = args.cif_file
    pdb_file = args.pdb_file
    
    doc = gemmi.cif.read_file(cif_file)
    block = doc.sole_block()  
    structure = gemmi.make_structure_from_block(block)
    structure.setup_entities()  
    with open(pdb_file, "w") as f:
        f.write(structure.make_pdb_string())
    
if __name__ == "__main__":

    argparser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    argparser.add_argument("--cif_file", type=str, help="Input CIF file.")
    argparser.add_argument("--pdb_file", type=str, help="Output PDB file.")

    args = argparser.parse_args()
    main(args)
""")

    # ------------------------------------------------------------------------------------------------------------------------------
    # Create the ESMfold.py script
    # ------------------------------------------------------------------------------------------------------------------------------
    with open(f'{self.FOLDER_PARENT}/ESMfold.py', 'w') as f:
        f.write("""
import argparse
import sys
from transformers import AutoTokenizer, EsmForProteinFolding, EsmConfig
import torch
from transformers.models.esm.openfold_utils.protein import to_pdb, Protein as OFProtein
from transformers.models.esm.openfold_utils.feats import atom14_to_atom37

def main(args):

    sequence_file = args.sequence_file
    output_file = args.output_file

    # Set PyTorch to use only one thread
    torch.set_num_threads(1)

    with open(sequence_file) as f: sequence=f.read()

    def convert_outputs_to_pdb(outputs):
        final_atom_positions = atom14_to_atom37(outputs["positions"][-1], outputs)
        outputs = {k: v.to("cpu").numpy() for k, v in outputs.items()}
        final_atom_positions = final_atom_positions.cpu().numpy()
        final_atom_mask = outputs["atom37_atom_exists"]
        pdbs = []
        for i in range(outputs["aatype"].shape[0]):
            aa = outputs["aatype"][i]
            pred_pos = final_atom_positions[i]
            mask = final_atom_mask[i]
            resid = outputs["residue_index"][i] + 1
            pred = OFProtein(
                aatype=aa,
                atom_positions=pred_pos,
                atom_mask=mask,
                residue_index=resid,
                b_factors=outputs["plddt"][i],
                chain_index=outputs["chain_index"][i] if "chain_index" in outputs else None,
            )
            pdbs.append(to_pdb(pred))
        return pdbs

    try:
        tokenizer = AutoTokenizer.from_pretrained("facebook/esmfold_v1")
        model = EsmForProteinFolding.from_pretrained("facebook/esmfold_v1", low_cpu_mem_usage=True)
        torch.backends.cuda.matmul.allow_tf32 = True
        model.trunk.set_chunk_size(64)

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")  
        model = model.to(device)
        tokenized_input = tokenizer([sequence], return_tensors="pt", add_special_tokens=False)['input_ids'].to(device)
        
        with torch.no_grad(): 
            output = model(tokenized_input)

        pdb = convert_outputs_to_pdb(output)
        with open(output_file, "w") as f: 
            f.write("".join(pdb))

    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        raise

    finally:
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            torch.cuda.synchronize()
    
if __name__ == "__main__":
    argparser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    argparser.add_argument("--sequence_file", type=str, help="File containing sequence to be predicted.")
    argparser.add_argument("--output_file", type=str, help="Output PDB.")

    args = argparser.parse_args()
    main(args)
""")


    # ------------------------------------------------------------------------------------------------------------------------------
    # Create the extract_sequence_from_pdb.py script
    # ------------------------------------------------------------------------------------------------------------------------------
    with open(f'{self.FOLDER_PARENT}/extract_sequence_from_pdb.py', 'w') as f:
        f.write("""
import argparse
import warnings
from Bio import BiopythonParserWarning
from Bio import SeqIO

def main(args):

    pdb_in = args.pdb_in
    sequence_out = args.sequence_out
         
    with open(pdb_in, "r") as f:
        for record in SeqIO.parse(f, "pdb-atom"):
            seq = str(record.seq)
    
    with open(sequence_out, "w") as f:
        f.write(seq)
 
if __name__ == "__main__":
    
    warnings.simplefilter("ignore", BiopythonParserWarning)
    
    argparser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    argparser.add_argument("--pdb_in", type=str, help="PDB file from which the sequence is read.")
    argparser.add_argument("--sequence_out", type=str, help="File into which the sequence is storred.")
     
    args = argparser.parse_args()
    main(args)
    
""")

    # ------------------------------------------------------------------------------------------------------------------------------
    # Create the find_highest_scoring_sequence.py script
    # ------------------------------------------------------------------------------------------------------------------------------
    with open(f'{self.FOLDER_PARENT}/find_highest_scoring_sequence.py', 'w') as f:
        f.write('''
import re
import argparse

def main(args):

    sequence_wildcard = args.sequence_wildcard
    sequence_parent   = args.sequence_parent
    sequence_in       = args.sequence_in
    sequence_out      = args.sequence_out
    
    # Read the parent sequence
    with open(sequence_parent, 'r') as file:
        sequence_parent = file.readline().strip()

    # Read the input sequence pattern and prepare it for regex matching
    with open(args.sequence_wildcard, 'r') as file:
        sequence_wildcard = file.readline().strip()
    sequence_wildcard = re.sub('X', '.', sequence_wildcard)  # Replace 'X' with regex wildcard '.'

    highest_score = 0
    highest_scoring_sequence = ''

    # Process the sequence file to find the highest scoring sequence
    with open(sequence_in, 'r') as file:
        for line in file:
            if line.startswith('>'):
                score_match = re.search('global_score=(\d+\.\d+)', line)
                if score_match:
                    score = float(score_match.group(1))
                    sequence = next(file, '').strip()  # Read the next line for the sequence
                    
                    # Check if the score is higher, the sequence is different from the parent,
                    # and does not match the input sequence pattern
                    if score > highest_score and sequence != sequence_parent and not re.match(sequence_wildcard, sequence):
                        highest_score = score
                        highest_scoring_sequence = sequence

    with open(sequence_out, 'w') as f:
        f.write(highest_scoring_sequence)

if __name__ == "__main__":
    argparser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    argparser.add_argument("--sequence_wildcard", type=str, help="Sequence file with wildcards for designable residues.")
    argparser.add_argument("--sequence_parent", type=str, help="Sequence file of design parent variant.")
    argparser.add_argument("--sequence_in", type=str, help="Sequence file containing all designed variants.")
    argparser.add_argument("--sequence_out", type=str, help="Output sequence file of best variant.")

    args = argparser.parse_args()
    main(args)

''')       
        


import os
import sys
import glob
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from PIL import Image
from IPython.display import display

from helper_002               import normalize_scores, hamming_distance, save_cat_res_into_all_scores_df, save_all_scores_df, exponential_func
from plotting_tree_001        import *

# Data loding function --------------------------------------------------------------------------------------------------------------------------

def make_plots(self):
    
    load_plot_scores_df(self)

    ### TEMPORARY WORK AROUND TO FACILITATE PLOTTING
    if not hasattr(self, "WEIGHT_TOTAL"): self.WEIGHT_TOTAL = 1.0
    if not hasattr(self, "WEIGHT_CATALYTIC"): self.WEIGHT_CATALYTIC = 1.0
    if not hasattr(self, "WEIGHT_INTERFACE"): self.WEIGHT_INTERFACE = 1.0
    if not hasattr(self, "WEIGHT_IDENTICAL"): self.WEIGHT_IDENTICAL = 1.0
    if not hasattr(self, "WEIGHT_EFIELD"): self.WEIGHT_EFIELD = 1.0
    if not hasattr(self, "WEIGHT_REDOX"): self.WEIGHT_REDOX = 1.0

    if self.PRINT_VALS:
        print_vals(self)
        
    if self.STATISTICS:
        plot_statistics(self)
        
    if self.SCORES_V_INDEX:
        plot_scores_v_index(self)
        
    if self.SCORES_V_GEN:
        plot_scores_v_generation(self)
        
    if self.SCORES_HIST:
        plot_scores_hist(self)
                    
    if self.PLOT_TREE:
        plot_tree(self,color=self.TREE_SCORE)

    if self.RESOURCE_LOG:
        plot_resource_log(self)
        
    if self.landscape_plot:
        landscape_plotting_function(self)

    combine_images(self)

def load_plot_scores_df(self):

    # Condition to check if the ALL_SCORES_CSV file exists, otherwise it returns the function.
    if not os.path.isfile(f'{self.FOLDER_HOME}/all_scores.csv'): 
        print(f"ERROR: {self.FOLDER_HOME}/all_scores.csv does not exist!")
        sys.exit()  

    # Load and modify scores_df
    self.all_scores_df = pd.read_csv(self.ALL_SCORES_CSV)
    self.plot_scores_df = self.all_scores_df.dropna(subset=['total_score'])
    
    # Check if there is enough data
    if len(self.plot_scores_df)<10:
        print(f"ERROR: {self.FOLDER_HOME}/all_scores.csv does not contain enough finished designs!")
        sys.exit()  
        
    # Calculate scores
    self.scores = normalize_scores(self, 
                                   self.plot_scores_df, 
                                   norm_all=True, 
                                   extension="score") 

    if "redox" in self.SELECTED_SCORES:
        self.scores['BioDC_redox'] = self.plot_scores_df["BioDC_redox"]
        self.HIGHSCORE['BioDC_redox'] = self.TARGET_REDOX

# Combined plot functions --------------------------------------------------------------------------------------------------------------------------

def plot_statistics(self):

    fig, axes = plt.subplots(1, 4, figsize=((len(self.SELECTED_SCORES) + 1) * self.PLOT_SIZE, self.PLOT_SIZE))
    
    # --- Chart 1: Completed vs. Not Finished ---
    finished_count = self.all_scores_df['total_score'].notna().sum()
    crashed_count = 0
    non_finished_count = 0

    na_rows = self.all_scores_df[self.all_scores_df["total_score"].isna()]
    for index, row in na_rows.iterrows():
        error_files = glob.glob(os.path.join(self.FOLDER_HOME, str(index), "scripts", "*.err"))
        has_error = False
        for err_file in error_files:
            with open(err_file, "r", errors="ignore") as f:
                if any("error" in line.lower() for line in f):
                    has_error = True
                    break 
        if not error_files or not has_error:
            non_finished_count += 1
        else:
            crashed_count += 1
            
    data = [finished_count, non_finished_count, crashed_count]
    labels = [f'Finished ({finished_count})', f'Unfinished ({non_finished_count})', f'Failed ({crashed_count})' ]
    colors = [(*color, 0.5) for color in reversed(plt.cm.Set1.colors[:3])]
    wedges, _ = axes[0].pie(data, autopct=None, colors=colors, startangle=90)
    axes[0].legend(wedges, labels, loc='upper center', bbox_to_anchor=(0.5, -0.05), ncol=1)
    axes[0].set_title("Completed")
    
    # --- Chart 2: Design Methods ---
    if 'design_method' in self.all_scores_df.columns:
        design_method_counts = self.all_scores_df['design_method'].value_counts()
        data = design_method_counts.values
        labels = [f'{label} ({count})' for label, count in zip(design_method_counts.index, design_method_counts.values)]
        colors = [(*color, 0.5) for color in plt.cm.Set1.colors[:len(design_method_counts)]]
        wedges, _ = axes[1].pie(data, autopct=None, colors=colors, startangle=90)
        axes[1].legend(wedges, labels, loc='upper center', bbox_to_anchor=(0.5, -0.05), ncol=1)
        axes[1].set_title("Methods")
    else:
        axes[1].set_visible(False)
    
    # --- Chart 3: Blocked (Running) ---
    if 'blocked' in self.all_scores_df.columns:
        blocked_counts = self.all_scores_df['blocked'].value_counts()
        # Exclude "unblocked"
        blocked_counts = blocked_counts[blocked_counts.index != "unblocked"]
        if not blocked_counts.empty:
            data = blocked_counts.values
            labels = [f'{label} ({count})' for label, count in zip(blocked_counts.index, blocked_counts.values)]
            colors = [(*color, 0.5) for color in plt.cm.Set1.colors[:len(blocked_counts)]]
            wedges, _ = axes[2].pie(data, autopct=None, colors=colors, startangle=90)
            axes[2].legend(wedges, labels, loc='upper center', bbox_to_anchor=(0.5, -0.05), ncol=1)
            axes[2].set_title("Running")
        else:
            axes[2].set_title("Nothing Running")
    else:
        axes[2].set_visible(False)
            
    # --- Chart 4: Next Steps ---
    if 'next_steps' in self.all_scores_df.columns:
        self.all_scores_df['next_step'] = self.all_scores_df['next_steps'].dropna().apply(lambda x: x.split(',')[0])
        next_step_counts = self.all_scores_df['next_step'].value_counts()
        data = next_step_counts.values
        labels = [f'{label} ({count})' for label, count in zip(next_step_counts.index, next_step_counts.values)]
        colors = [(*color, 0.5) for color in plt.cm.Set1.colors[:len(next_step_counts)]]
        wedges, _ = axes[3].pie(data, autopct=None, colors=colors, startangle=90)
        axes[3].legend(wedges, labels, loc='upper center', bbox_to_anchor=(0.5, -0.05), ncol=1)
        axes[3].set_title("Next Step")
    else:
        axes[3].set_visible(False)
    
    plt.tight_layout()
    save_path = os.path.join(self.FOLDER_PLOT, "plot_statistics.png")
    fig.savefig(save_path, bbox_inches='tight')
    plt.close(fig)

def plot_scores_hist(self):
    
    fig, axes = plt.subplots(1, len(self.SELECTED_SCORES) + 1,
                             figsize=((len(self.SELECTED_SCORES) + 1) * self.PLOT_SIZE, self.PLOT_SIZE))
    
    for idx, ax in enumerate(axes):
        
        if idx == 0:
            score_type = "combined"
        else:
            score_type = self.SELECTED_SCORES[idx - 1]
            
        plot_score_hist(self, ax, f'{score_type}_score')
        
    plt.tight_layout()
    save_path = os.path.join(self.FOLDER_PLOT, "plot_scores_hist.png")
    fig.savefig(save_path, bbox_inches='tight')
    plt.close(fig)

def plot_scores_v_index(self):
    
    fig, axes = plt.subplots(1, len(self.SELECTED_SCORES) + 1,
                             figsize=((len(self.SELECTED_SCORES) + 1) * self.PLOT_SIZE, self.PLOT_SIZE))
    
    for idx, ax in enumerate(axes):
        
        if idx == 0:
            score_type = "final"
        else:
            score_type = self.SELECTED_SCORES[idx - 1]
            
        if score_type in ["redox"]:
            score_plotted = "BioDC_redox"
        else:
            score_plotted = f'{score_type}_score'
            
        plot_score_v_index(self, ax, score_plotted)
        
    plt.tight_layout()
    save_path = os.path.join(self.FOLDER_PLOT, "plot_scores_v_index.png")
    fig.savefig(save_path, bbox_inches='tight')
    plt.close(fig)

def plot_scores_v_generation(self):
    
    fig, axes = plt.subplots(1, len(self.SELECTED_SCORES) + 1,
                             figsize=((len(self.SELECTED_SCORES) + 1) * self.PLOT_SIZE, self.PLOT_SIZE))
    
    for idx, ax in enumerate(axes):
        if idx == 0:
            score_type = "final"
        else:
            score_type = self.SELECTED_SCORES[idx - 1]
            
        if score_type in ["redox"]:
            score_plotted = "BioDC_redox"
        else:
            score_plotted = f'{score_type}_score'
            
        plot_score_v_generation_violin(self, ax, score_plotted)
        
    plt.tight_layout()
    save_path = os.path.join(self.FOLDER_PLOT, "plot_scores_v_generation.png")
    fig.savefig(save_path, bbox_inches='tight')
    plt.close(fig)

def plot_mutations_v_generation(self):
    
    fig, axes = plt.subplots(1, len(self.SELECTED_SCORES) + 1,
                             figsize=((len(self.SELECTED_SCORES) + 1) * self.PLOT_SIZE, self.PLOT_SIZE))

    plot_mutations_v_generation_violin(self, axes[0])
    plot_hamming_distance_v_generation_violin(self, axes[1])
    for idx in range(2,len(self.SELECTED_SCORES) + 1, 1):
        axes[idx].axis('off')
    
    plt.tight_layout()
    save_path = os.path.join(self.FOLDER_PLOT, "plot_mutations_v_generation.png")
    fig.savefig(save_path, bbox_inches='tight')
    plt.close(fig)
    
def print_vals(self):
    rows = []
    for score_type in [score for score in self.scores if "score" in score]:

        row = {
            "score_type": score_type,
            "min": min(self.scores[score_type]),
            "max": max(self.scores[score_type]),
            "highscore": self.HIGHSCORE.get(f'{score_type}', None),
            "negbest": self.NEGBEST.get(f'{score_type}', None)
        }
        if "combined" not in score_type:
            row["norm_min"] = self.NORM[score_type][0]
            row["norm_max"] = self.NORM[score_type][1]
        else:
            row["norm_min"] = None
            row["norm_max"] = None
        rows.append(row)

    df = pd.DataFrame(rows)
    df["min"] = df["min"].map(lambda x: f"{x:.2f}")
    df["max"] = df["max"].map(lambda x: f"{x:.2f}")
    df = df[["score_type","min", "max", "norm_min", "norm_max", "highscore", "negbest"]]
    
    # Render the DataFrame as a table using matplotlib
    fig, ax = plt.subplots(figsize=((len(self.SELECTED_SCORES) + 1) * self.PLOT_SIZE, len(rows) * 0.5 + 1))
    ax.axis('tight')
    ax.axis('off')
    ax.table(cellText=df.values, colLabels=df.columns, loc='center')
    plt.tight_layout()
    save_path = os.path.join(self.FOLDER_PLOT, "print_vals.png")
    fig.savefig(save_path, bbox_inches='tight')
    plt.close(fig)

def combine_images(self):
    # List the file names you saved from the above functions
    file_names = ["print_vals.png" ]

    if self.STATISTICS:     file_names += [ "plot_statistics.png"]
    if self.SCORES_HIST:    file_names += [ "plot_scores_hist.png"]
    if self.SCORES_V_INDEX: file_names += [ "plot_scores_v_index.png"]
    if self.SCORES_V_GEN:   file_names += [ "plot_scores_v_generation.png"]

    file_paths = [os.path.join(self.FOLDER_PLOT, name) for name in file_names]

    # Open images using PIL
    images = [Image.open(fp) for fp in file_paths if os.path.exists(fp)]
    if not images:
        print("No images found to combine.")
        return

    # Compute the total size (vertical stacking)
    widths, heights = zip(*(im.size for im in images))
    total_width = max(widths)
    total_height = sum(heights)

    # Create a new blank image with white background
    combined_image = Image.new('RGB', (total_width, total_height), color=(255, 255, 255))
    
    # Paste each image into the combined image
    y_offset = 0
    for im in images:
        combined_image.paste(im, (0, y_offset))
        y_offset += im.size[1]
    
    # Save the combined image
    combined_path = os.path.join(self.FOLDER_PLOT, "results.png")
    combined_image.save(combined_path)
    display(combined_image)

def plot_resource_log(self):

    # Load perfomrance data
    self.resource_log_df = pd.read_csv(self.RESOURCE_LOG_CSV)

    # Compute the relative time (in seconds) based on the first row
    time = self.resource_log_df['time'] - self.resource_log_df['time'].iloc[0]

    # Create 3 subplots
    fig, axes = plt.subplots(1, 3, figsize=(3 * self.PLOT_SIZE, self.PLOT_SIZE))
    
    # --- Plot 1: CPU and GPU usage as percentages ---
    # CPU usage percentage (divide by self.MAX_JOBS)
    cpu_usage_pct = self.resource_log_df['cpus_used'] / self.MAX_JOBS * 100

    # GPU usage percentage (divide by self.MAX_GPUS, if any GPUs available)
    if self.MAX_GPUS > 0:
        gpu_usage_pct = self.resource_log_df['gpus_used'] / self.MAX_GPUS * 100
    else:
        gpu_usage_pct = [0] * len(self.resource_log_df)

    axes[0].plot(time, cpu_usage_pct, label='CPU', marker='o')
    axes[0].plot(time, gpu_usage_pct, label='GPU', marker='o')
    axes[0].set_ylabel("Usage (%)")
    axes[0].set_title("CPU and GPU Usage")
    axes[0].set_ylim(0,100)
    axes[0].set_xlim(left=0)
    axes[0].legend()

    # --- Plot 2: Design count ---
    total_designs = self.resource_log_df['total_designs']
    unfinished_designs = self.resource_log_df['unfinished_designs']
    finished_designs = self.resource_log_df['finished_designs']
    failed_designs = self.resource_log_df['failed_designs']
    axes[1].plot(time, total_designs, label='Total', marker='o')
    axes[1].plot(time, finished_designs, label='Finished', marker='o')
    axes[1].plot(time, unfinished_designs, label='Unfinished', marker='o')
    axes[1].plot(time, failed_designs, label='Failed', marker='o')
    axes[1].set_ylabel("Number of Designs")
    axes[1].set_xlabel("Relative Time (seconds)")
    axes[1].set_title("Total Designs")
    axes[1].set_ylim(bottom=0)
    axes[1].set_xlim(left=0)
    axes[1].legend()

    plt.tight_layout()
    save_path = os.path.join(self.FOLDER_PLOT, "resource_log_performance.png")
    fig.savefig(save_path, bbox_inches='tight')
    plt.close(fig)
    
# Sub-plot functions --------------------------------------------------------------------------------------------------------------------------

def plot_score_v_index(self, ax, score_type):

    # Prepares data for the plot
    normalized_scores = self.scores[score_type]
    
    # Plot score
    ax.scatter(self.plot_scores_df.index, normalized_scores, c='lightgrey', s=5) 
    moving_avg = pd.Series(normalized_scores).rolling(window=20).mean()
    ax.plot(range(len(moving_avg)),moving_avg,c="k")

    if score_type == "BioDC_redox":
        ylim = (-1,1)
    else:
        ylim = (0,1)

    # Adds the two values HIGHSCORE and NEG_BEST as reference values for the top ever design and the best ever negative control value.
    if score_type in self.HIGHSCORE: ax.axhline(self.HIGHSCORE[score_type], color='b', label='Highest Score', alpha=0.5)
    if score_type in self.NEGBEST:   ax.axhline(self.NEGBEST[score_type], color='r', label='Negative Best', alpha=0.5)

    # Sets the plot details
    ax.set_title(score_type.replace("_", " "))
    ax.set_xlabel('index')
    ax.set_ylim(ylim)
    ax.set_xlim(left=0)
    if score_type not in ["BioDC_redox"]:
        ax.tick_params(axis='y', which='both', left=False, labelleft=False)
 
def plot_score_hist(self, ax, score_type):

    # Prepares data for the plot
    normalized_scores = self.scores[score_type]

    xmin = -0.1
    xmax = 1.1
    
    # Plot score
    ax.hist(normalized_scores, bins=np.arange(xmin,xmax,0.02),color='grey')

    if score_type == "combined_score":

        # Calculate current kbt value        
        generation=self.plot_scores_df['generation'].max()
        if isinstance(self.KBT_BOLTZMANN, (float, int)):
            kbt_boltzmann = self.KBT_BOLTZMANN
        elif len(self.KBT_BOLTZMANN) == 2:
            kbt_boltzmann = self.KBT_BOLTZMANN[0] * np.exp(-self.KBT_BOLTZMANN[1]*generation)
        elif len(self.KBT_BOLTZMANN) == 3:
            kbt_boltzmann = (self.KBT_BOLTZMANN[0]-self.KBT_BOLTZMANN[2])*np.exp(-self.KBT_BOLTZMANN[1]*generation)+self.KBT_BOLTZMANN[2]

        """ THAT WAS A BAD IDEA! CAN BE REMOVED LIKELY SOON!!!
        normalized_scores = np.sort(normalized_scores)[-1000:]
        """
        
        boltzmann_factors = np.exp(normalized_scores / kbt_boltzmann)
        probabilities = boltzmann_factors / sum(boltzmann_factors)
        boltzmann_scores = np.random.choice(normalized_scores, size=100000, replace=True, p=probabilities)
        ax_dup = ax.twinx()
        ax_dup.hist(boltzmann_scores, bins=np.arange(-0.1,1.1,0.02), density=True, alpha=0.7, color='orange', label=f'kbt = {kbt_boltzmann:.2f}')
        ax_dup.tick_params(axis='y', which='both', right=False, labelright=False)
        ax_dup.legend()  

    # Adds the two values HIGHSCORE and NEG_BEST as reference values for the top ever design and the best ever negative control value.
    if score_type in self.HIGHSCORE: ax.axvline(self.HIGHSCORE[score_type], color='b', label='Highest Score', alpha=0.5)
    if score_type in self.NEGBEST:   ax.axvline(self.NEGBEST[score_type], color='r', label='Negative Best', alpha=0.5)

    # Sets the plot details
    ax.set_title(score_type.replace("_", " "))
    ax.set_xlabel(score_type)
    ax.set_ylim(bottom=0)
    ax.set_xlim(0,1)
    ax.tick_params(axis='y', which='both', left=False, labelleft=False)
    
def plot_score_v_generation_violin(self, ax, score_type):
    
    # Prepares data for the plot
    normalized_scores = self.scores[score_type]
    generations = self.plot_scores_df['generation']

    # Group scores by generation using an explicit for loop.
    scores_v_generations = []
    for generation in range(0, max(generations)+1, 1):
        current_gen_scores = [] 
        for score, gen in zip(normalized_scores, generations):
            if gen == generation:
                current_gen_scores.append(score)
        scores_v_generations.append(current_gen_scores)
        
    # Plot score
    parts = ax.violinplot(scores_v_generations, positions=range(0,max(generations)+1,1))

    # Customizing the color of violin plots 
    for pc in parts['bodies']:
        pc.set_facecolor('grey')
        pc.set_edgecolor('none')
        pc.set_alpha(0.5)
    for partname in ('cbars', 'cmins', 'cmaxes'):
        vp = parts.get(partname)
        vp.set_visible(False)
            
    # Add mean points
    means = [np.mean(data) for data in scores_v_generations]  
    inds = np.arange(0, len(means), 1)  
    ax.scatter(inds, means, marker='o', color='k', s=20, zorder=3)  

    if score_type ==  "BioDC_redox":
        ylim = (-1,1)
    else:
        ylim = (0,1)

    #Adds the two values HIGHSCORE and NEG_BEST as reference values for the top ever design and the best ever negative control value.
    if score_type in self.HIGHSCORE: ax.axhline(self.HIGHSCORE[score_type], color='b', label='Highest Score', alpha=0.5)
    if score_type in self.NEGBEST:   ax.axhline(self.NEGBEST[score_type], color='r', label='Negative Best', alpha=0.5)

    #Sets the plot details
    ax.set_title(score_type.replace("_", " "))
    ax.set_xlabel('generation')
    ax.set_ylim(ylim)
    ax.set_xlim(left=-0.3)
    ax.set_xticks(range(0, max(generations)+1))
    if score_type not in ["BioDC_redox"]:
        ax.tick_params(axis='y', which='both', left=False, labelleft=False)


import os
import subprocess
import logging
import re
import json
import shutil
import argparse
import pandas as pd
import numpy as np
from Bio import SeqIO

def run_command(command, cwd=None, capture_output=False):
    """Wrapper to execute .py files in runtime with arguments, and print error messages if they occur.
    
    Parameters:
    - command: The command to run as a list of strings.
    - cwd: Optional; The directory to execute the command in.
    - capture_output: Optional; If True, capture stdout and stderr. Defaults to False (This is to conserve memory).
    """
    try:
        # If capture_output is True, capture stdout and stderr
        if capture_output:
            result = subprocess.run(command, capture_output=True, text=True, check=True, cwd=cwd)
        else:
            # If capture_output is False, suppress all output by redirecting to os.devnull
            with open(os.devnull, 'w') as fnull:
                result = subprocess.run(command, stdout=fnull, stderr=fnull, text=True, check=True, cwd=cwd)
        return result.stdout
    except subprocess.CalledProcessError as e:
        logging.error(f"Command '{e.cmd}' failed with return code {e.returncode}")
        logging.error(e.stderr)
        raise
    except Exception as e:
        logging.error(f"An error occurred while running command: {command}")
        raise

def submit_job(index, job, bash=False, ram=16, variables_json=None):

    if variables_json:
        FOLDER_HOME = variables_json['FOLDER_HOME']
        SUBMIT_PREFIX = variables_json['SUBMIT_PREFIX']
        ABBIE_LOCAL = variables_json['ABBIE_LOCAL']
        BLUEPEBBLE = variables_json['BLUEPEBBLE']
        BACKGROUND_JOB = variables_json['BACKGROUND_JOB']
        GRID = variables_json['GRID']

    if GRID:
        submission_script = f"""#!/bin/bash
#$ -V
#$ -cwd
#$ -N {SUBMIT_PREFIX}_{job}_{index}
#$ -hard -l mf={ram}G
#$ -o {FOLDER_HOME}/{index}/scripts/AI_{job}_{index}.out
#$ -e {FOLDER_HOME}/{index}/scripts/AI_{job}_{index}.err
"""
    if BLUEPEBBLE:
        submission_script = f"""#!/bin/bash
#SBATCH --account={BLUEPEBBLE_ACCOUNT}
#SBATCH --partition=short
#SBATCH --mem=40GB
#SBATCH --ntasks-per-node=1
#SBATCH --time=2:00:00    
#SBATCH --nodes=1          
#SBATCH --job-name={SUBMIT_PREFIX}_{job}_{index}
#SBATCH --output={FOLDER_HOME}/{index}/scripts/AI_{job}_{index}.out
#SBATCH --error={FOLDER_HOME}/{index}/scripts/AI_{job}_{index}.err
"""
        
    if BACKGROUND_JOB:
        if not os.path.isfile(f'{FOLDER_HOME}/n_running_jobs.dat'):
            with open(f'{FOLDER_HOME}/n_running_jobs.dat', 'w') as f: f.write('0')
        with open(f'{FOLDER_HOME}/n_running_jobs.dat', 'r'): jobs = int(f.read())
        with open(f'{FOLDER_HOME}/n_running_jobs.dat', 'w'): f.write(jobs+1)
        submission_script = ""

    if ABBIE_LOCAL:
        submission_script = ""
        
    submission_script += f"""
# Output folder
cd {FOLDER_HOME}/{index}
pwd
bash {FOLDER_HOME}/{index}/scripts/{job}_{index}.sh
""" 
    if BACKGROUND_JOB:
        submission_script = f"""
jobs=$(cat {FOLDER_HOME}/n_running_jobs.dat)
jobs=$((jobs - 1))
echo "$jobs" > {FOLDER_HOME}/n_running_jobs.dat
"""


    # Create the submission_script
    with open(f'{FOLDER_HOME}/{index}/scripts/submit_{job}_{index}.sh', 'w') as file: file.write(submission_script)
    
    if bash:
        #Bash the submission_script for testing
        subprocess.run(f'bash {FOLDER_HOME}/{index}/scripts/submit_{job}_{index}.sh', shell=True, text=True)
    else:
        #Submit the submission_script
        if GRID:
            if "ESM" in job:
                
                output = subprocess.check_output(
    (f'qsub -l h="!bs-dsvr64&!bs-dsvr58&!bs-dsvr42&'
     f'!bs-grid64&!bs-grid65&!bs-grid66&!bs-grid67&'
     f'!bs-grid68&!bs-grid69&!bs-grid70&!bs-grid71&'
     f'!bs-grid72&!bs-grid73&!bs-grid74&!bs-grid75&'
     f'!bs-grid76&!bs-grid77&!bs-grid78&!bs-headnode04&'
     f'!bs-stellcontrol05&!bs-stellsubmit05" -q regular.q '
     f'{FOLDER_HOME}/{index}/scripts/submit_{job}_{index}.sh'),
    shell=True, text=True
    )
            else:
                output = subprocess.check_output(f'qsub -q regular.q \
                                                {FOLDER_HOME}/{index}/scripts/submit_{job}_{index}.sh', \
                                                shell=True, text=True)
            logging.debug(output[:-1]) #remove newline at end of output
            
        if BLUEPEBBLE:
            output = subprocess.check_output(f'sbatch {FOLDER_HOME}/{index}/scripts/submit_{job}_{index}.sh', \
                                             shell=True, text=True)
            logging.debug(output[:-1]) #remove newline at end of output
            
        if BACKGROUND_JOB:

            stdout_log_file_path = f'{FOLDER_HOME}/{index}/scripts/submit_{job}_{index}_stdout.log'
            stderr_log_file_path = f'{FOLDER_HOME}/{index}/scripts/submit_{job}_{index}_stderr.log'

            with open(stdout_log_file_path, 'w') as stdout_log_file, open(stderr_log_file_path, 'w') as stderr_log_file:
                process = subprocess.Popen(f'bash {FOLDER_HOME}/{index}/scripts/submit_{job}_{index}.sh &', 
                                           shell=True, stdout=stdout_log_file, stderr=stderr_log_file)
        
        if ABBIE_LOCAL:

            stdout_log_file_path = f'{FOLDER_HOME}/{index}/scripts/submit_{job}_{index}_stdout.log'
            stderr_log_file_path = f'{FOLDER_HOME}/{index}/scripts/submit_{job}_{index}_stderr.log'

            with open(stdout_log_file_path, 'w') as stdout_log_file, open(stderr_log_file_path, 'w') as stderr_log_file:
                process = subprocess.Popen(f'bash {FOLDER_HOME}/{index}/scripts/submit_{job}_{index}.sh &', 
                                           shell=True, stdout=stdout_log_file, stderr=stderr_log_file)

def extract_sequence_from_pdb(pdb_path):
    with open(pdb_path, "r") as pdb_file:
        for record in SeqIO.parse(pdb_file, "pdb-atom"):
            seq = str(record.seq)
    return seq

def find_highest_scoring_sequence(folder_path, parent_index, input_sequence_path, variables_json):
    """
    Identifies the highest scoring protein sequence from a set of generated PMPNN sequences,
    excluding the parent and WT sequence (except wildcard positions specified by DESIGN).

    Parameters:
    - folder_path (str): The path to the directory containing sequence files (/ProteinMPNN).
    - parent_index (str): The index of the parent protein sequence.
    - input_sequence_path (str): The path to a file containing the input sequence pattern,
      where 'X' represents wildcard positions that can match any character.
    - -------------------------------- GLOBAL variables used ----------------------------
    - WT (str): The wild type or reference protein identifier.
      

    Returns:
    - highest_scoring_sequence (str): The protein sequence with the highest score 
      that does not match the parent and WT.
    
    Note:
    This function parses .fa files to find sequences and their scores, and applies
    a regex pattern derived from the input sequence to filter sequences.
    It assumes the presence of 'global_score' within the sequence descriptor lines
    in the .fa file for scoring.
    """

    WT = variables_json['WT']
    # Construct the file path for the sequence data
    file_path = f'{folder_path}/seqs/{WT}_Rosetta_Relax_{parent_index}.fa'
    parent_seq_file = f'{folder_path}/Rosetta_Relax_{parent_index}.seq'
    
    # Read the parent sequence from its file
    with open(parent_seq_file, 'r') as file:
        parent_sequence = file.readline().strip()

    # Read the input sequence pattern and prepare it for regex matching
    with open(input_sequence_path, 'r') as file:
        input_sequence = file.readline().strip()
    pattern = re.sub('X', '.', input_sequence)  # Replace 'X' with regex wildcard '.'

    highest_score = 0
    highest_scoring_sequence = ''

    # Process the sequence file to find the highest scoring sequence
    with open(file_path, 'r') as file:
        for line in file:
            if line.startswith('>'):
                score_match = re.search('global_score=(\d+\.\d+)', line)
                if score_match:
                    score = float(score_match.group(1))
                    sequence = next(file, '').strip()  # Read the next line for the sequence
                    
                    # Check if the score is higher, the sequence is different from the parent,
                    # and does not match the input sequence pattern
                    if score > highest_score and sequence != parent_sequence and not re.match(pattern, sequence):
                        highest_score = score
                        highest_scoring_sequence = sequence

    # Return the highest scoring sequence found
    return highest_scoring_sequence

def save_cat_res_into_all_scores_df(all_scores_df, index, PDB_file_path, from_parent_struct=False):
    
    '''Finds the indices and names of the catalytic residue from <PDB_file_path> 
       Saves indices and residues into <all_scores_df> in row <index> as lists.
       To make sure these are saved and loaded as list, ";".join() and .split(";") should be used
       IF information is read from an input structure for design do not save cat_resn
       Returns the updated all_scores_df'''
    
    with open(PDB_file_path, 'r') as f: 
        PDB = f.readlines()
    
    remarks = [i for i in PDB if i[:10] == 'REMARK 666']

    cat_resis = []
    cat_resns = []

    for remark in remarks:
        cat_resis.append(str(int(remark[55:59])))

    for cat_resi in cat_resis:
        for line in PDB[len(remarks)+2:]:
            atomtype = line[12:16]
            if atomtype != " CA ": continue
            resi = str(int(line[22:26]))
            resn = line[17:20]
            if resi == cat_resi:
                cat_resns.append(resn)
                break

    all_scores_df.at[index, 'cat_resi'] = ";".join(cat_resis)
    # Only save the cat_resn if this comes from the designed structure, not from the input structure for design
    if not from_parent_struct:
        all_scores_df.at[index, 'cat_resn'] = ";".join(cat_resns)
    
    
    return all_scores_df

def generate_remark_from_all_scores_df(all_scores_df, index, variables_json):

    LIGAND = variables_json['LIGAND']

    remark = ''
    cat_resns = str(all_scores_df.at[index, 'cat_resn']).split(';')
    # Making sure resi is converted to int to avoid crash in Relax
    cat_resis = [int(float(x)) for x in str(all_scores_df.at[index, 'cat_resi']).split(';')]
    
    remarks = []

    for idx, (cat_resi, cat_resn) in enumerate(zip(cat_resis, cat_resns), start=1):
        remarks.append(f'REMARK 666 MATCH TEMPLATE X {LIGAND}    0 MATCH MOTIF A {cat_resn}{str(cat_resi).rjust(5)}  {idx}  1')
    return "\n".join(remarks)

def run_ProteinMPNN(parent_index, new_index, all_scores_df, variables_json):
    """
    Executes the ProteinMPNN pipeline for a given protein structure and generates
    new protein sequences with potentially higher functional scores. Also calls 
    ESMfold_RosettaRelax with the ProteinMPNN flag set to True.

    Parameters:
    - parent_index (str): The index of the parent protein variant.
    - new_index (str): The index assigned to the new protein variant.
    - all_scores_df (DataFrame): A DataFrame containing information for protein variants.
    - --------------------------------GLOBAL variables used (from variables.json) ----------------------------
    - FOLDER_HOME (str): The base directory where ProteinMPNN and related files are located.
    - WT (str): The wild type or reference protein identifier.
    - DESIGN (str): A string representing positions and types of amino acids to design with RosettaDesign.
    - ProteinMPNN_T (float): The sampling temperature for ProteinMPNN.
    - EXPLORE (bool): Flag to indicate whether exploration mode is enabled.
    - PMPNN_BIAS (float): The bias value for ProteinMPNN parent sequence retention.

    Returns:
    None: The function writes the highes scoring sequence to the specified directories.
    
    Note:
    This function assumes the ProteinMPNN toolkit is available and properly set up in the specified location.
    It involves multiple subprocess calls to Python scripts for processing protein structures and generating new sequences.
    """
    # Set GLOBALS
    FOLDER_HOME = variables_json['FOLDER_HOME']
    WT = variables_json['WT']
    DESIGN = variables_json['DESIGN']
    ProteinMPNN_T = variables_json['ProteinMPNN_T']
    EXPLORE = variables_json['EXPLORE']
    PMPNN_BIAS = variables_json['PMPNN_BIAS']

    # Ensure ProteinMPNN is available
    if not os.path.exists(f'{FOLDER_HOME}/../ProteinMPNN'):
        logging.error(f"ProteinMPNN not installed in {FOLDER_HOME}/../ProteinMPNN.")
        logging.error("Install using: git clone https://github.com/dauparas/ProteinMPNN.git")
        return

    # Prepare file paths
    pdb_file = f"{FOLDER_HOME}/{parent_index}/{WT}_Rosetta_Relax_{parent_index}.pdb"
    if not os.path.isfile(pdb_file):
        logging.error(f"{pdb_file} not present!")
        return

    protein_mpnn_folder = f"{FOLDER_HOME}/{new_index}/ProteinMPNN"
    os.makedirs(protein_mpnn_folder, exist_ok=True)
    shutil.copy(pdb_file, os.path.join(protein_mpnn_folder, f"{WT}_Rosetta_Relax_{parent_index}.pdb"))

    seq = extract_sequence_from_pdb(pdb_file)
    with open(os.path.join(protein_mpnn_folder, f"Rosetta_Relax_{parent_index}.seq"), "w") as f:
        f.write(seq)
    

    # Run ProteinMPNN steps using subprocess after creating the bias file
    helper_scripts_path = f"{FOLDER_HOME}/../ProteinMPNN/helper_scripts"
    protein_mpnn_path = f"{FOLDER_HOME}/../ProteinMPNN"

    # Prepare input JSON for bias dictionary creation
    input_json = {"name": f"{WT}_Rosetta_Relax_{parent_index}", "seq_chain_A": seq}

    # Create bias dictionary
    mpnn_alphabet = 'ACDEFGHIKLMNPQRSTVWYX'
    mpnn_alphabet_dict = {aa: idx for idx, aa in enumerate(mpnn_alphabet)}
    
    bias_dict = {}
    for chain_key, sequence in input_json.items():
        if chain_key.startswith('seq_chain_'):
            chain = chain_key[-1]
            chain_length = len(sequence)
            bias_per_residue = np.zeros([chain_length, 21])  # 21 for each amino acid in the alphabet

            # Apply a positive bias for the amino acid at each position
            for idx, aa in enumerate(sequence):
                if aa in mpnn_alphabet_dict:  # Ensure the amino acid is in the defined alphabet
                    aa_index = mpnn_alphabet_dict[aa]
                    bias_per_residue[idx, aa_index] = PMPNN_BIAS  # Use the global bias variable

            bias_dict[input_json["name"]] = {chain: bias_per_residue.tolist()}

    # Write the bias dictionary to a JSON file
    bias_json_path = os.path.join(protein_mpnn_folder, "bias_by_res.jsonl")
    with open(bias_json_path, 'w') as f:
        json.dump(bias_dict, f)
        f.write('\n')

    # Parse multiple chains
    run_command([
        "python", os.path.join(helper_scripts_path, "parse_multiple_chains.py"),
        "--input_path", protein_mpnn_folder,
        "--output_path", os.path.join(protein_mpnn_folder, "parsed_chains.jsonl")
    ])

    # Assign fixed chains
    run_command([
        "python", os.path.join(helper_scripts_path, "assign_fixed_chains.py"),
        "--input_path", os.path.join(protein_mpnn_folder, "parsed_chains.jsonl"),
        "--output_path", os.path.join(protein_mpnn_folder, "assigned_chains.jsonl"),
        "--chain_list", 'A'
    ])

    # Make fixed positions dict
    run_command([
        "python", os.path.join(helper_scripts_path, "make_fixed_positions_dict.py"),
        "--input_path", os.path.join(protein_mpnn_folder, "parsed_chains.jsonl"),
        "--output_path", os.path.join(protein_mpnn_folder, "fixed_positions.jsonl"),
        "--chain_list", 'A',
        "--position_list", " ".join(DESIGN.split(","))
    ])

    # Protein MPNN run
    run_command([
        "python", os.path.join(protein_mpnn_path, "protein_mpnn_run.py"),
        "--jsonl_path", os.path.join(protein_mpnn_folder, "parsed_chains.jsonl"),
        "--chain_id_jsonl", os.path.join(protein_mpnn_folder, "assigned_chains.jsonl"),
        "--fixed_positions_jsonl", os.path.join(protein_mpnn_folder, "fixed_positions.jsonl"),
        "--bias_by_res_jsonl", os.path.join(protein_mpnn_folder, "bias_by_res.jsonl"),
        "--out_folder", protein_mpnn_folder,
        "--num_seq_per_target", "100",
        "--sampling_temp", ProteinMPNN_T,
        "--seed", "37",
        "--batch_size", "1"
    ])
    

    # Find highest scoring sequence
    highest_scoring_sequence = find_highest_scoring_sequence(protein_mpnn_folder, parent_index, input_sequence_path=f"{FOLDER_HOME}/input_sequence_with_X_as_wildecard.seq", variables_json=variables_json)

    # Save highest scoring sequence and prepare for ESMfold
    with open(os.path.join(protein_mpnn_folder, f"{WT}_{new_index}.seq"), "w") as f:
        f.write(highest_scoring_sequence)
    
    if highest_scoring_sequence:
        logging.info(f"Ran ProteinMPNN for index {parent_index} and found a new sequence with index {new_index}.")
    else:
        logging.error(f"Failed to find a new sequnce for index {parent_index} with ProteinMPNN.")
    
    all_scores_df = save_cat_res_into_all_scores_df(all_scores_df, new_index, pdb_file, from_parent_struct=False)
        
    # Run ESMfold Relax with the ProteinMPNN Flag
    return run_ESMfold_RosettaRelax(index=new_index, all_scores_df=all_scores_df, OnlyRelax=False, \
                             ProteinMPNN=True, ProteinMPNN_parent_index=parent_index, bash=True, variables_json=variables_json)

def run_ESMfold_RosettaRelax(index, all_scores_df, OnlyRelax=False, ProteinMPNN=False, PreMatchRelax=False,
                             ProteinMPNN_parent_index=0, cmd="", bash=False, EXPLORE=False, variables_json=None):
    
    # Set GLOBALS
    FOLDER_HOME = variables_json['FOLDER_HOME']
    FOLDER_INPUT = variables_json['FOLDER_INPUT']
    WT = variables_json['WT']
    DESIGN = variables_json['DESIGN']
    GRID = variables_json['GRID']
    BLUEPEBBLE = variables_json['BLUEPEBBLE']
    BACKGROUND_JOB = variables_json['BACKGROUND_JOB']
    ABBIE_LOCAL = variables_json['ABBIE_LOCAL']
    ROSETTA_PATH = variables_json['ROSETTA_PATH']
    LIGAND = variables_json['LIGAND']

    # Giving the ESMfold algorihm the needed inputs
    output_file = f'{FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_output_{index}.pdb'

    protein_mpnn_seq_file = f'{FOLDER_HOME}/{index}/ProteinMPNN/{WT}_{index}.seq'

    if os.path.exists(protein_mpnn_seq_file):
        sequence_file = protein_mpnn_seq_file
    else:
        logging.error(f"{protein_mpnn_seq_file} not present, but PMPNN was run prior to ESMfold.")
        return False


        
    # Make directories
    os.makedirs(f"{FOLDER_HOME}/{index}/ESMfold", exist_ok=True)
    os.makedirs(f"{FOLDER_HOME}/{index}/scripts", exist_ok=True)

        
    # Options for EXPLORE, accelerated script for testing
    ex = "-ex1 -ex2"
    if EXPLORE: ex = ""
        
    # Get Name of parent PDB
    PDBFile = f"{FOLDER_HOME}/{ProteinMPNN_parent_index}/{WT}_Rosetta_Relax_{ProteinMPNN_parent_index}.pdb"

    if not os.path.isfile(PDBFile):
        logging.error(f"{PDBFile} not present!")
        return False
                    
    # Get the pdb file from the last step and strip away ligand and hydrogens 
    cpptraj = f'''parm    {PDBFile}
trajin  {PDBFile}
strip   :{LIGAND}
strip   !@C,N,O,CA
trajout {FOLDER_HOME}/{index}/ESMfold/{WT}_CPPTraj_Apo_{index}.pdb
'''
    with open(f'{FOLDER_HOME}/{index}/ESMfold/CPPTraj_Apo_{index}.in','w') as f: f.write(cpptraj)

    # Get the pdb file from the last step and strip away everything except the ligand
    cpptraj = f'''parm    {PDBFile}
trajin  {PDBFile}
strip   !:{LIGAND}
trajout {FOLDER_HOME}/{index}/ESMfold/{WT}_CPPTraj_Lig_{index}.pdb
'''
    with open(f'{FOLDER_HOME}/{index}/ESMfold/CPPTraj_Lig_{index}.in','w') as f: f.write(cpptraj)

    # Get the ESMfold pdb file and strip away all hydrogens
    cpptraj = f'''parm    {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_output_{index}.pdb
trajin  {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_output_{index}.pdb
strip   !@C,N,O,CA
trajout {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_no_hydrogens_{index}.pdb
'''
    with open(f'{FOLDER_HOME}/{index}/ESMfold/CPPTraj_no_hydrogens_{index}.in','w') as f: f.write(cpptraj)

    # Align substrate and ESM prediction of scaffold without hydrogens
    cpptraj = f'''parm    {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_no_hydrogens_{index}.pdb
reference {FOLDER_HOME}/{index}/ESMfold/{WT}_CPPTraj_Apo_{index}.pdb [apo]
trajin    {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_no_hydrogens_{index}.pdb
rmsd      @CA ref [apo]
trajout   {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_aligned_{index}.pdb noter
'''
    with open(f'{FOLDER_HOME}/{index}/ESMfold/CPPTraj_aligned_{index}.in','w') as f: f.write(cpptraj) 
              
    if GRID:           extension = "linuxgccrelease"
    if BLUEPEBBLE:     extension = "serialization.linuxgccrelease"
    if BACKGROUND_JOB: extension = "serialization.linuxgccrelease"
    if ABBIE_LOCAL:    
        extension = "linuxgccrelease"
        bash_args = "OMP_NUM_THREADS=1"
    else:
        bash_args = ""
 
    cmd += f"""
    
{bash_args} python {FOLDER_HOME}/ESMfold.py {output_file} {sequence_file}

sed -i '/PARENT N\/A/d' {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_output_{index}.pdb
cpptraj -i {FOLDER_HOME}/{index}/ESMfold/CPPTraj_Apo_{index}.in           &> \
           {FOLDER_HOME}/{index}/ESMfold/CPPTraj_Apo_{index}.out
cpptraj -i {FOLDER_HOME}/{index}/ESMfold/CPPTraj_Lig_{index}.in           &> \
           {FOLDER_HOME}/{index}/ESMfold/CPPTraj_Lig_{index}.out
cpptraj -i {FOLDER_HOME}/{index}/ESMfold/CPPTraj_no_hydrogens_{index}.in  &> \
           {FOLDER_HOME}/{index}/ESMfold/CPPTraj_no_hydrogens_{index}.out
cpptraj -i {FOLDER_HOME}/{index}/ESMfold/CPPTraj_aligned_{index}.in       &> \
           {FOLDER_HOME}/{index}/ESMfold/CPPTraj_aligned_{index}.out

# Assemble the final protein
sed -i '/END/d' {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_aligned_{index}.pdb
# Return HETATM to ligand output and remove TER
sed -i -e 's/^ATOM  /HETATM/' -e '/^TER/d' {FOLDER_HOME}/{index}/ESMfold/{WT}_CPPTraj_Lig_{index}.pdb
"""
    
    input_extension_relax = ""
    if PreMatchRelax:
        extension_relax = "_APO"
        ## No ligand necessary so just use the aligned pdb from ESMfold
        cmd += f"""
cp {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_aligned_{index}.pdb \
   {FOLDER_HOME}/{index}/{WT}_ESMfold_{index}{extension_relax}.pdb
"""  

    else:
        extension_relax = ""
        remark = generate_remark_from_all_scores_df(all_scores_df, index, variables_json)
        with open(f'{FOLDER_HOME}/{index}/{WT}_ESMfold_{index}.pdb', 'w') as f: f.write(remark+"\n")
        cmd += f"""
cat {FOLDER_HOME}/{index}/ESMfold/{WT}_ESMfold_aligned_{index}.pdb >> {FOLDER_HOME}/{index}/{WT}_ESMfold_{index}.pdb
cat {FOLDER_HOME}/{index}/ESMfold/{WT}_CPPTraj_Lig_{index}.pdb     >> {FOLDER_HOME}/{index}/{WT}_ESMfold_{index}.pdb
sed -i '/TER/d' {FOLDER_HOME}/{index}/{WT}_ESMfold_{index}.pdb
"""
        
    cmd += f"""
# Run Rosetta Relax
{ROSETTA_PATH}/bin/rosetta_scripts.{extension} \
                -s                                        {FOLDER_HOME}/{index}/{WT}_ESMfold_{index}{extension_relax}.pdb \
                -extra_res_fa                             {FOLDER_INPUT}/{LIGAND}.params \
                -parser:protocol                          {FOLDER_HOME}/{index}/scripts/Rosetta_Relax_{index}.xml \
                -out:file:scorefile                       {FOLDER_HOME}/{index}/score_rosetta_relax.sc \
                -nstruct                                  1 \
                -ignore_zero_occupancy                    false \
                -corrections::beta_nov16                  true \
                -run:preserve_header                      true \
                -overwrite {ex}

# Rename the output file
mv {WT}_ESMfold_{index}{extension_relax}_0001.pdb {WT}_Rosetta_Relax_{index}{extension_relax}.pdb
sed -i '/        H  /d' {WT}_Rosetta_Relax_{index}{extension_relax}.pdb
"""
    
    if PreMatchRelax:
        extension_relax = "_APO"
        
        cmd += f"""
# Align relaxed ESM prediction of scaffold without hydrogens
cpptraj -i {FOLDER_HOME}/{index}/ESMfold/{WT}_Rosetta_Relax_aligned_{index}{extension_relax}.in           &> \
           {FOLDER_HOME}/{index}/ESMfold/{WT}_Rosetta_Relax_aligned_{index}{extension_relax}.out
sed -i '/END/d' {FOLDER_HOME}/{index}/{WT}_Rosetta_Relax_aligned_{index}{extension_relax}.pdb
"""  
        
        cpptraj = f'''parm    {FOLDER_HOME}/{index}/{WT}_Rosetta_Relax_{index}{extension_relax}.pdb [protein]
parm      {FOLDER_HOME}/{index}/ESMfold/{WT}_CPPTraj_Apo_{index}.pdb [reference]
reference {FOLDER_HOME}/{index}/ESMfold/{WT}_CPPTraj_Apo_{index}.pdb parm [reference] [apo]
trajin    {FOLDER_HOME}/{index}/{WT}_Rosetta_Relax_{index}{extension_relax}.pdb parm [protein]
rmsd      @CA ref [apo]
trajout   {FOLDER_HOME}/{index}/{WT}_Rosetta_Relax_aligned_{index}{extension_relax}.pdb noter
'''
        with open(f'{FOLDER_HOME}/{index}/ESMfold/{WT}_Rosetta_Relax_aligned_{index}{extension_relax}.in','w') as f: 
            f.write(cpptraj) 
    
        # Create the Rosetta_Relax.xml file
    repeats = "3"
    if EXPLORE: repeats = "1"
    Rosetta_Relax_xml = f"""
<ROSETTASCRIPTS>

    <SCOREFXNS>
    
        <ScoreFunction name      = "score"                   weights = "beta_nov16" >
            <Reweight scoretype  = "atom_pair_constraint"    weight  = "1" />
            <Reweight scoretype  = "angle_constraint"        weight  = "1" />
            <Reweight scoretype  = "dihedral_constraint"     weight  = "1" />
        </ScoreFunction> 
        
        <ScoreFunction name      = "score_final"             weights = "beta_nov16" >
            <Reweight scoretype  = "atom_pair_constraint"    weight  = "1" />
            <Reweight scoretype  = "angle_constraint"        weight  = "1" />
            <Reweight scoretype  = "dihedral_constraint"     weight  = "1" />
        </ScoreFunction>
        
    </SCOREFXNS>
       
    <MOVERS>
                                  
        <FastRelax  name="mv_relax" disable_design="false" repeats="{repeats}" /> 
"""
    if not PreMatchRelax: Rosetta_Relax_xml += f"""
        <AddOrRemoveMatchCsts     name="mv_add_cst" 
                                  cst_instruction="add_new" 
                                  cstfile="{FOLDER_INPUT}/{LIGAND}_{WT}_enzdes_planar.cst" />

"""
    Rosetta_Relax_xml += f"""

        <InterfaceScoreCalculator   name                   = "mv_inter" 
                                    chains                 = "X" 
                                    scorefxn               = "score_final" />
    </MOVERS>
    
    <PROTOCOLS>  

        <Add mover_name="mv_relax" />
"""
    if not PreMatchRelax: Rosetta_Relax_xml += f"""                                  
        <Add mover_name="mv_add_cst" />       
        <Add mover_name="mv_inter" />
"""
    Rosetta_Relax_xml += f"""
    </PROTOCOLS>
    
</ROSETTASCRIPTS>
"""
    # Write the Rosetta_Relax.xml to a file
    with open(f'{FOLDER_HOME}/{index}/scripts/Rosetta_Relax_{index}.xml', 'w') as f:
        f.writelines(Rosetta_Relax_xml)      
        
    if ProteinMPNN:
        with open(f'{FOLDER_HOME}/{index}/scripts/ProteinMPNN_ESMfold_Rosetta_Relax_{index}.sh', 'w') as file:
            file.write(cmd)
        logging.info(f"Run Ligand/ProteinMPNN for index {index} based on index {ProteinMPNN_parent_index}.")
        submit_job(index=index, job="ProteinMPNN_ESMfold_Rosetta_Relax", bash=bash, variables_json=variables_json)

    return True

def main():
    # Parse command line arguments
    parser = argparse.ArgumentParser(description="Run ProteinMPNN and ESMfold Rosetta Relax workflows.")
    parser.add_argument("--index", type=int, required=True, help="Index for the current run.")
    parser.add_argument("--new_index", type=int, required=True, help="New index for the current run.")
    parser.add_argument("--home_folder", type=str, required=True, help="Path to the home folder.")
    args = parser.parse_args()

    # Read the variables.json file
    variables_json = f'{args.home_folder}/variables.json'
    with open(variables_json, 'r') as file:
        variables = json.load(file)
    
    #add home folder to variables
    variables['FOLDER_HOME'] = args.home_folder
    variables['FOLDER_INPUT'] = os.path.join(os.path.dirname(variables['FOLDER_HOME']), 'Input')

    all_scores_df = pd.read_csv(f'{args.home_folder}/all_scores.csv')

    log_format = '%(asctime)s - %(levelname)s - %(message)s'
    date_format = '%Y-%m-%d %H:%M:%S'
    logging.basicConfig(filename=f'{variables["FOLDER_HOME"]}.log', level=logging.DEBUG, format=log_format, datefmt=date_format)

    # Run the ProteinMPNN process
    success = run_ProteinMPNN(parent_index=args.index, new_index=args.new_index, all_scores_df=all_scores_df, variables_json=variables)

    if success:
        logging.info(f"Process completed successfully for index {args.index}.")
    else:
        logging.error(f"Process failed for index {args.index}.")

if __name__ == "__main__":
    main()



import os
import subprocess
import numpy as np
import pandas as pd
import pickle as pkl

from helper_002               import *

def prepare_efields(self, index:str, cmd:str):
    """
    Calculate electric fields for structure in {index}.
    
    Paramters:
    index (str): The index of the protein variant for which efields are to be calculated.
    cmd (str): Growing list of commands to be exected by run_design using submit_job.

    Returns:
    cmd (str): Command to be exected by run_design using submit_job.
    """
    
    # Define files and make empty directory
    filename_in = self.all_scores_df.at[index, "final_variant"]
    filename_out = f'{self.FOLDER_DESIGN}/{index}/ElectricFields/{os.path.basename(filename_in)}'
    os.makedirs(f'{self.FOLDER_DESIGN}/{index}/ElectricFields', exist_ok=True)
    
    # Make tleap files to generate input
    with open(f"{filename_out}_tleap.in", "w") as f:
        f.write(f"""source leaprc.protein.ff19SB 
    source leaprc.gaff
    loadamberprep   {self.FOLDER_INPUT}/{self.LIGAND}.prepi
    loadamberparams {self.FOLDER_INPUT}/{self.LIGAND}.frcmod
    mol = loadpdb {filename_out}.pdb
    saveamberparm mol {filename_out}.parm7 {filename_out}.rst7
    quit
    """)
            
    cmd += f"""### ElectricFields ###

# Delete everything after 'CONECT' and remove hydrogens 
sed -n '/CONECT/q;p' {filename_in}.pdb > \\
                     {filename_out}.pdb
sed -i '/ H /d'      {filename_out}.pdb

# Make AMBER files
tleap -f             {filename_out}_tleap.in > \\
                     {filename_out}_tleap.out
mv leap.log          {filename_out}_tleap.log

# Add field calculation command
python   {self.FIELD_TOOLS} \\
-nc      {filename_out}.rst7 \\
-parm    {filename_out}.parm7 \\
-out     {filename_out}_fields.pkl \\
-target  {self.FOLDER_PARENT}/field_target.dat \\
-solvent WAT
"""
    
    if self.REMOVE_TMP_FILES:
        cmd += f"""# Removing temporary directory
rm -r ESMFold
"""
   
    return cmd

def get_efields_score(self, index, score_type):

    with open(f"{self.FOLDER_DESIGN}/{index}/ElectricFields/{self.WT}_{score_type}_{index}_fields.pkl", "rb") as f:
        FIELDS = pkl.load(f)

    bond_field =  np.array(FIELDS[self.FIELD_TARGET.replace(" ", "_")]['Total'])
    
    if self.FIELDS_EXCL_CAT:
        key = f"{self.all_scores_df.at[index, 'cat_resn']}_{self.all_scores_df.at[index, 'cat_resi']}"
        bond_field = bond_field - np.array(FIELDS[self.FIELD_TARGET.replace(" ", "_")][key])
        #print(f'original field {bond_field}')
        #print(f'base field {FIELDS[self.FIELD_TARGET.replace(" ", "_")][key]}')
        #print(f'new field base subtraction {bond_field}')
        
    all_fields = FIELDS[self.FIELD_TARGET.replace(" ", "_")]

    return bond_field[0], all_fields

def update_efieldsdf(self, index:int, index_efields_dict:dict):
    '''Adds a new row to "{FOLDER_HOME}/electric_fields.csv" containing the electric fields 
    generated by FieldTools.py for all residues in the protein'''

    no_residues = len(index_efields_dict)-4

    gen_headers = ["Total","Protein","Solvent","WAT"]
    resi_headers = [f"RESI_{idx}" for idx in range(1,no_residues+1)]
    headers = gen_headers + resi_headers

    fields_list = [field[0] for field in index_efields_dict.values()]

    if not os.path.isfile(f"{self.FOLDER_HOME}/electric_fields.csv"):       
        fields_df = pd.DataFrame([fields_list], columns=headers, index=[index])
        fields_df.to_csv(f"{self.FOLDER_HOME}/electric_fields.csv") 

    else:
        fields_df = pd.read_csv(f"{self.FOLDER_HOME}/electric_fields.csv", index_col=0)
        new_row_df = pd.DataFrame([fields_list], columns=headers, index=[index])
        fields_df = pd.concat([fields_df, new_row_df])
        fields_df.sort_index(inplace=True)
        fields_df.to_csv(f"{self.FOLDER_HOME}/electric_fields.csv") 



"""
Manages all MPNN methods for protein design

Functions:
    prepare_ProteinMPNN():  Prepares commands for ProteinMPNN job submission.
    prepare_SolubleMPNN():  Prepares commands for SolubleMPNN job submission.
    prepare_LigandMPNN():   Prepares commands for LigandMPNN job submission.
    make_bias_dict():       Creates and writes a bias dictionary for MPNN based on an input PDB.
    ProteinMPNN_check():    Checks that ProteinMPNN is installed.
    LigandMPNN_check():     Checks that LigandMPNN is installed.

Modules Required:
    helper_002
"""
import os
import logging
import json
import numpy as np
import sys

from helper_002               import sequence_from_pdb

def prepare_LigandMPNN(self, 
                       index, 
                       cmd,
                       gpu_id = None):
    """
    Uses LigandMPNN to redesign the input structure with index.

    Parameters:
        index (str):    The index of the designed variant.
        cmd (str):      Growing list of commands to be exected by run_design using submit_job.
    Optional Parameters:
        gpu_id (int):   ID of the GPU to be used.
    Returns:
        cmd (str):      Command to be exected by run_design using submit_job
    """
    
    LigandMPNN_check(self)
        
    folder_ligandmpnn = f"{self.FOLDER_DESIGN}/{index}/LigandMPNN"
    os.makedirs(folder_ligandmpnn, exist_ok=True)
    
    PDB_input = self.all_scores_df.at[int(index), "step_input_variant"]
    if "parent" in PDB_input:
        seq_input = PDB_input
    else:
        seq_input = "_".join(PDB_input.split("_")[:-2]+PDB_input.split("_")[-1:])
        
    make_bias_dict(self, PDB_input, folder_ligandmpnn)

    cmd = f"""### LigandMPNN ###
"""

    if  gpu_id != None:
        cmd += f"""
export CUDA_VISIBLE_DEVICES={gpu_id}
"""
        
    cmd += f"""    
# Copy input PDB
cp {PDB_input}.pdb {folder_ligandmpnn}

echo 'XXXXXXXXXXXXXXXXXXXXXX Extract catalytic residue information -- TO DO!!!! FIX RESIDUES!!!'

# Extract catalytic residue information -- TO DO!!!! FIX RESIDUES!!!
# cat_resi = int(all_scores_df.at[parent_index, 'cat_resi'])
fixed_residues = f"A1"

echo 'XXXXXXXXXXXXXXXXXXXXXX ADD --use_soluble_model TAG???'
echo 'XXXXXXXXXXXXXXXXXXXXXX ADD --bias_by_res_jsonl {os.path.join(folder_solublempnn, 'bias_by_res.jsonl')}'

# Run LigandMPNN
{self.bash_args} python {os.path.join(self.FOLDER_LigandMPNN, 'run.py')} \
--model_type ligand_mpnn \
--temperature {self.LigandMPNN_T} \
--seed 37 \
--pdb_path os.path.join(ligand_mpnn_folder, f"{WT}_Rosetta_Relax_{parent_index}.pdb"),
--out_folder {folder_ligandmpnn} \
--pack_side_chains 1 \
--number_of_packs_per_design 100 \
--fixed_residues fixed_residues

# Find highest scoring sequence
{self.bash_args} python {os.path.join(self.FOLDER_PARENT, 'find_highest_scoring_sequence.py')} \
--sequence_wildcard {self.FOLDER_HOME}/{self.WT}_with_X_as_wildecard.seq \
--sequence_parent   {seq_input}.seq \
--sequence_in       {folder_ligandmpnn}/seqs/{os.path.splitext(os.path.basename(PDB_input))[0]}.fa \
--sequence_out      {self.WT}_{index}.seq 
"""                              
    
    if self.REMOVE_TMP_FILES:
        cmd += f"""
# Removing temporary directory
rm -r LigandMPNN
"""
        
    return(cmd)
    
def prepare_ProteinMPNN(self,
                        index, 
                        cmd,
                        gpu_id = None):
    """
    Uses ProteinMPNN to redesign the input structure with index.

    Args:
        index (str): The index of the designed variant.
        cmd (str): Growing list of commands to be exected by run_design using submit_job.

    Returns:
        cmd (str): Command to be exected by run_design using submit_job
    """
    
    ProteinMPNN_check(self)
        
    folder_proteinmpnn = f"{self.FOLDER_DESIGN}/{index}/ProteinMPNN"
    os.makedirs(folder_proteinmpnn, exist_ok=True)
    
    PDB_input = self.all_scores_df.at[int(index), "step_input_variant"]
    if "parent" in PDB_input:
        seq_input = PDB_input
    else:
        seq_input = "_".join(PDB_input.split("_")[:-2]+PDB_input.split("_")[-1:])
    
    make_bias_dict(self, PDB_input, folder_proteinmpnn)

    cmd = f"""### ProteinMPNN ###
"""

    if  gpu_id != None:
        cmd += f"""
export CUDA_VISIBLE_DEVICES={gpu_id}
"""
        
    cmd += f"""    
# Copy input PDB
cp {PDB_input}.pdb {folder_proteinmpnn}

# Parse chains
{self.bash_args} python {os.path.join(self.FOLDER_ProteinMPNN_h, 'parse_multiple_chains.py')} \
--input_path {folder_proteinmpnn} \
--output_path {os.path.join(folder_proteinmpnn, 'parsed_chains.jsonl')} 

# Assign fixed chains
{self.bash_args} python {os.path.join(self.FOLDER_ProteinMPNN_h, 'assign_fixed_chains.py')} \
--input_path {os.path.join(folder_proteinmpnn, 'parsed_chains.jsonl')} \
--output_path {os.path.join(folder_proteinmpnn, 'assigned_chains.jsonl')} \
--chain_list A 

# Make fixed positions dict
{self.bash_args} python {os.path.join(self.FOLDER_ProteinMPNN_h, 'make_fixed_positions_dict.py')} \
--input_path {os.path.join(folder_proteinmpnn, 'parsed_chains.jsonl')} \
--output_path {os.path.join(folder_proteinmpnn, 'fixed_positions.jsonl')} \
--chain_list A \
--position_list '{" ".join(self.DESIGN.split(","))}' 

# Protein MPNN run
{self.bash_args} python {os.path.join(self.FOLDER_ProteinMPNN, 'protein_mpnn_run.py')} \
--jsonl_path {os.path.join(folder_proteinmpnn, 'parsed_chains.jsonl')} \
--chain_id_jsonl {os.path.join(folder_proteinmpnn, 'assigned_chains.jsonl')} \
--fixed_positions_jsonl {os.path.join(folder_proteinmpnn, 'fixed_positions.jsonl')} \
--bias_by_res_jsonl {os.path.join(folder_proteinmpnn, 'bias_by_res.jsonl')} \
--out_folder {folder_proteinmpnn} \
--num_seq_per_target 100 \
--sampling_temp {self.ProteinMPNN_T} \
--seed 37 \
--batch_size 1

# Find highest scoring sequence
{self.bash_args} python {os.path.join(self.FOLDER_PARENT, 'find_highest_scoring_sequence.py')} \
--sequence_wildcard {self.FOLDER_HOME}/{self.WT}_with_X_as_wildecard.seq \
--sequence_parent   {seq_input}.seq \
--sequence_in       {folder_proteinmpnn}/seqs/{os.path.splitext(os.path.basename(PDB_input))[0]}.fa \
--sequence_out      {self.WT}_{index}.seq 

"""                              
    
    return(cmd)

def prepare_SolubleMPNN(self, 
                        index,
                        cmd,
                        gpu_id = None):
    """
    Uses SolubleMPNN to redesign the input structure with index.

    Args:
        index (str): The index of the designed variant.
        cmd (str): Growing list of commands to be exected by run_design using submit_job.

    Returns:
        cmd (str): Command to be exected by run_design using submit_job
    """
    
    ProteinMPNN_check(self)
        
    folder_solublempnn = f"{self.FOLDER_DESIGN}/{index}/SolubleMPNN"
    os.makedirs(folder_solublempnn, exist_ok=True)
    
    PDB_input = self.all_scores_df.at[int(index), "step_input_variant"]
    if "parent" in PDB_input:
        seq_input = PDB_input
    else:
        seq_input = "_".join(PDB_input.split("_")[:-2]+PDB_input.split("_")[-1:])
        
    make_bias_dict(self, PDB_input, folder_solublempnn)

    cmd = f"""### SolubleMPNN ###
"""

    if  gpu_id != None:
        cmd += f"""
export CUDA_VISIBLE_DEVICES={gpu_id}
"""
        
    cmd += f"""    
# Copy input PDB
cp {PDB_input}.pdb {folder_solublempnn}

# Parse chains
{self.bash_args} python {os.path.join(self.FOLDER_ProteinMPNN_h, 'parse_multiple_chains.py')} \
--input_path {folder_solublempnn} \
--output_path {os.path.join(folder_solublempnn, 'parsed_chains.jsonl')} 

# Assign fixed chains
{self.bash_args} python {os.path.join(self.FOLDER_ProteinMPNN_h, 'assign_fixed_chains.py')} \
--input_path {os.path.join(folder_solublempnn, 'parsed_chains.jsonl')} \
--output_path {os.path.join(folder_solublempnn, 'assigned_chains.jsonl')} \
--chain_list A 

# Make fixed positions dict
{self.bash_args} python {os.path.join(self.FOLDER_ProteinMPNN_h, 'make_fixed_positions_dict.py')} \
--input_path {os.path.join(folder_solublempnn, 'parsed_chains.jsonl')} \
--output_path {os.path.join(folder_solublempnn, 'fixed_positions.jsonl')} \
--chain_list A \
--position_list '{" ".join(self.DESIGN.split(","))}' 

# SolubleMPNN run
{self.bash_args} python {os.path.join(self.FOLDER_ProteinMPNN, 'protein_mpnn_run.py')} \
--use_soluble_model \
--jsonl_path {os.path.join(folder_solublempnn, 'parsed_chains.jsonl')} \
--chain_id_jsonl {os.path.join(folder_solublempnn, 'assigned_chains.jsonl')} \
--fixed_positions_jsonl {os.path.join(folder_solublempnn, 'fixed_positions.jsonl')} \
--bias_by_res_jsonl {os.path.join(folder_solublempnn, 'bias_by_res.jsonl')} \
--out_folder {folder_solublempnn} \
--num_seq_per_target 100 \
--sampling_temp {self.SolubleMPNN_T} \
--seed 37 \
--batch_size 1

# Find highest scoring sequence
{self.bash_args} python {os.path.join(self.FOLDER_PARENT, 'find_highest_scoring_sequence.py')} \
--sequence_wildcard {self.FOLDER_HOME}/{self.WT}_with_X_as_wildecard.seq \
--sequence_parent   {seq_input}.seq \
--sequence_in       {folder_solublempnn}/seqs/{os.path.splitext(os.path.basename(PDB_input))[0]}.fa \
--sequence_out      {self.WT}_{index}.seq 

"""                              
    
    return(cmd)
    
def make_bias_dict(self, PDB_input, folder_mpnn):
    """
    Creates a bias dictionary for the input PDB and writes it to a JSON file.
    
    Args:
        PDB_input (str): Path to the input PDB file.
        folder_mpnn (str): Folder where the bias JSON will be saved.
    """ 
    
    # Prepare input JSON for bias dictionary creation
    seq = sequence_from_pdb(PDB_input)
    input_json = {"name": f"{os.path.basename(PDB_input)}", "seq_chain_A": seq}

    # Create bias dictionary
    mpnn_alphabet = 'ACDEFGHIKLMNPQRSTVWYX'
    mpnn_alphabet_dict = {aa: residue for residue, aa in enumerate(mpnn_alphabet)}
    
    bias_dict = {}
    for chain_key, sequence in input_json.items():
        if chain_key.startswith('seq_chain_'):
            chain = chain_key[-1]
            chain_length = len(sequence)
            bias_per_residue = np.zeros([chain_length, 21])  # 21 for each amino acid in the alphabet

            # Apply a positive bias for the amino acid at each position
            for idx, aa in enumerate(sequence):

                # Ensure the amino acid is in the defined alphabet
                if aa not in mpnn_alphabet_dict:  
                    logging.error(f"Non-standard amino acid {aa} at residue {idx} in make_bias_dict() / designMPNN.py.")
                    sys.exit()
                    
                # Forbid cysteine
                cysteine_index = mpnn_alphabet_dict['C']
                bias_per_residue[idx, cysteine_index] = -1e6
                
                # Add biase to input sequence
                aa_index = mpnn_alphabet_dict[aa]
                if "Soluble" in folder_mpnn:
                    bias_per_residue[idx, aa_index] = self.SolubleMPNN_BIAS 
                elif "Protein" in folder_mpnn:
                    bias_per_residue[idx, aa_index] = self.ProteinMPNN_BIAS 
                else:
                    logging.error(f"MPNN_BIAS not defined. Error in make_bias_dict() / designMPNN.py.")
                    sys.exit()  
                
            # Restrict aminoacids for restricted set
            if self.RESTRICT_RESIDUES is not None:
                for idx, resns in self.RESTRICT_RESIDUES:
                    for aa_index, aa in enumerate(mpnn_alphabet):
                        if aa in resns:
                            bias_per_residue[idx-1, aa_index] = 1
                        else:
                            bias_per_residue[idx-1, aa_index] = -1
                    
            bias_dict[input_json["name"]] = {chain: bias_per_residue.tolist()}

    # Write the bias dictionary to a JSON file
    bias_json_path = os.path.join(folder_mpnn, "bias_by_res.jsonl")
    with open(bias_json_path, 'w') as f:
        json.dump(bias_dict, f)
        f.write('\n')       
        
def ProteinMPNN_check(self):
    """
    Checks if ProteinMPNN is installed
    """
    if not os.path.exists(self.FOLDER_ProteinMPNN):
        logging.error(f"ProteinMPNN not installed in {self.FOLDER_ProteinMPNN}.")
        logging.error(f"Install using: 'cd {self.FOLDER_ProteinMPNN} && git clone https://github.com/dauparas/ProteinMPNN.git'")
        sys.exit()
        
def LigandMPNN_check(self):
    """
    Checks if LigandMPNN is installed
    """
    if not os.path.exists(self.FOLDER_LigandMPNN):
        logging.error(f"LigandMPNN not installed in {self.FOLDER_LigandMPNN}.")
        logging.error(f"Install using: 'cd {self.FOLDER_LigandMPNN} && git clone https://github.com/dauparas/LigandMPNN.git'")
        sys.exit()
    else:
        logging.error("LigandMPNN NOT WORKING YET!!! Problem with introducing biasedict!!!")
        sys.exit()
        

    


"""
Prepares commands to run MD minimization using Amber to refine protein structures.

Functions:
    - prepare_MDMin(): Sets up commands for sidechain building (Rosetta), generating Amber input files (tleap), and MD minimization (sander)
"""
import logging
import os
import shutil 
import subprocess  

from helper_002 import generate_remark_from_all_scores_df
              
def prepare_MDMin(self, 
                  index,  
                  cmd,
                  ):
    """
    Minimises protein structure in {index} using Amber.
    
    Parameters:
    - index (str): The index of the protein variant to be relaxed.
    - cmd (str): collection of commands to be run, this script wil append its commands to cmd
    """
    
    filename = f'{self.FOLDER_DESIGN}/{index}'
        
    # Make directories
    os.makedirs(f"{filename}/scripts", exist_ok=True)
    os.makedirs(f"{filename}/MDMin", exist_ok=True)

    PDBfile_in = self.all_scores_df.at[int(index), "step_input_variant"]    
    working_dir_path = f"{filename}/MDMin/{self.WT}_{index}"

    cmd += f"""### MDMin ###
    
# Assemble structure
cat {PDBfile_in}.pdb > {working_dir_path}_input.pdb
"""
    
# Write Rosetta input to build in sidechains
    Rosetta_xml = f"""
<ROSETTASCRIPTS>
    <MOVERS>
        <DumpPdb name="output_pdb" fname="{working_dir_path}_wRosetta_sidechains.pdb"/>
    </MOVERS>
    <PROTOCOLS>
        <Add mover_name="output_pdb" />
    </PROTOCOLS>
</ROSETTASCRIPTS>
"""
    with open(f'{filename}/scripts/Rosetta_buildPDB_{index}.xml', 'w') as f:
        f.write(Rosetta_xml)

# Add Rosetta command to shell script
    cmd += f"""
# Run Rosetta Relax
{self.ROSETTA_PATH}/bin/rosetta_scripts.{self.rosetta_ext} \\
    -s                        {working_dir_path}_input.pdb """
    if self.LIGAND not in ['HEM']:
        cmd += f"""\\
    -extra_res_fa                             {self.FOLDER_INPUT}/{self.LIGAND}.params """
    cmd += f"""\\
    -parser:protocol          {filename}/scripts/Rosetta_buildPDB_{index}.xml \\
    -nstruct                  1 \\
    -ignore_zero_occupancy    false \\
    -corrections::beta_nov16  true \\
    -run:preserve_header      true \\
                    
# Tidy up the output file
sed -i -e '/[0-9]H/d' -e '/H[0-9]/d' -e '/CONECT/Q' {working_dir_path}_wRosetta_sidechains.pdb

#Remove Rosetta output 
rm {self.WT}_{index}_input_0001.pdb

#Clean PDB for amber
pdb4amber -i {working_dir_path}_wRosetta_sidechains.pdb -o {working_dir_path}_clean4amber.pdb 2> {working_dir_path}_pdb4amber.log
rm {working_dir_path}_clean4amber_*

"""

    tleap = f"""source leaprc.protein.ff19SB 
source leaprc.gaff"""
    if self.LIGAND not in ['HEM']:
        tleap += f"""
loadamberprep   {self.FOLDER_INPUT}/{self.LIGAND}.prepi
loadamberparams {self.FOLDER_INPUT}/{self.LIGAND}.frcmod
"""
    tleap += f"""        
mol = loadpdb {working_dir_path}_clean4amber.pdb
set default pbradii mbondi3
saveamberparm mol {working_dir_path}.parm7 {working_dir_path}.rst7
quit
"""

    with open(f"{working_dir_path}_tleap.in", "w") as f:
        f.write(tleap)

    cmd += f"""# Generate AMBER input files
tleap -s -f {working_dir_path}_tleap.in &> {working_dir_path}_tleap.out
"""

    # Create the input file    
    in_file = f"""
initial minimization
&cntrl                                                                         
    imin           = 1,                                                        
    ntmin          = 0,                                                        
    ncyc           = 500,                                                      
    maxcyc         = 1000,                                                      
    ntpr           = 100,                                                                                                                   
    ntb=0                                                                      
    igb = 8                                                                    
    cut = 1000,                                                                                                                   
&end 
"""    
    # Write the Amber.in to a file
    with open(f'{filename}/scripts/MDmin_{index}.in', 'w') as f:
        f.writelines(in_file)  

    cmd += f"""
# Run MD
echo Starting MD
sander -O -i {filename}/scripts/MDmin_{index}.in \
          -c {working_dir_path}.rst7 -p {working_dir_path}.parm7 \
          -o {working_dir_path}_MD.log -x {working_dir_path}_MD.nc -r {working_dir_path}_MD_out.rst7 -inf {working_dir_path}_MD_out.mdinf

cpptraj -p {working_dir_path}.parm7 -y {working_dir_path}_MD_out.rst7 -x {working_dir_path}_MD_out.pdb

# Clean the output file
sed -i '/        H  /d' {working_dir_path}_MD_out.pdb
    """
    
    if self.CST_NAME is not None:
        remark = generate_remark_from_all_scores_df(self, index)
        cmd += f"""
echo '{remark}' > {self.WT}_MDMin_{index}.pdb"""
    
    cmd += f"""
cat {working_dir_path}_MD_out.pdb >> {self.WT}_MDMin_{index}.pdb
sed -i -e 's/^\(ATOM.\{{17\}}\) /\\1A/' {self.WT}_MDMin_{index}.pdb
sed -i -e 's/{self.LIGAND} A/{self.LIGAND} X/g' -e 's/HIE/HIS/g' -e 's/HID/HIS/g' {self.WT}_MDMin_{index}.pdb
"""
        
    if self.REMOVE_TMP_FILES:
        cmd += f"""
# Removing temporary directory
rm -r MDMin
"""

    return cmd



"""
Main Design Module. Coordinates various design steps, managing the workflow of Rosetta, ProteinMPNN, and other modules
within the AIzymes project.

Functions:
    get_ram: Determines RAM allocation for design steps.
    run_design: Runs the selected design steps based on configuration.

Modules Required:
    helper_002, design_match_001, design_ProteinMPNN_001, design_LigandMPNN_001,
    design_RosettaDesign_001, design_ESMfold_001, design_RosettaRelax_001
"""
import logging
import sys

from helper_002               import *

from design_match_001         import *
from design_MPNN_001          import *
from design_RosettaDesign_001 import *
from design_ESMfold_001       import *
from design_RosettaRelax_001  import *
from scoring_efields_001      import *   
from design_MDMin_001         import *  
from design_AlphaFold3_001    import *   
from scoring_BioDC_001        import *      

def get_ram(design_steps):
    
    ram = 0
    
    for design_step in design_steps:
        
        if design_step == "ProteinMPNN":
            new_ram = 20
        elif design_step == "SolubleMPNN":
            new_ram = 20
        elif design_step == "LigandMPNN":
            new_ram = 20
        elif design_step == "AlphaFold3MSA":
            new_ram = 20
        elif design_step == "AlphaFold3INF":
            new_ram = 20
        elif design_step == "RosettaDesign":
            new_ram = 10
        elif design_step == "RosettaRelax": 
            new_ram = 10
        elif design_step == "MDMin": 
            new_ram = 10
        elif design_step == "BioDC": 
            new_ram = 10
        elif design_step == "ESMfold":
            new_ram = 40
        elif design_step == "ElectricFields":
            new_ram = 10
        else:
            logging.error(f"RAM for design_step {design_step} is not defined in get_ram() / main_design.py")
            sys.exit()
            
        if new_ram > ram: ram = new_ram
            
    return ram

def run_design(self, 
               index,
               design_steps :list,
               bash = False
              ):
    
    # Expecting list. To make sure individual commmand would also be accepted, convert string to list if string is given
    if not isinstance(design_steps, list):
        variable = [design_steps]
    
    ram = get_ram(design_steps)
    
    cmd = ""
     
    for design_step in design_steps:

        # Assign GPU!
        gpu_id = None
        if self.MAX_GPUS > 0:
            if design_step in self.SYS_GPU_METHODS:
                for idx, job in self.gpus.items():
                    if job is None: 
                        gpu_id = idx
                        break
                if gpu_id == None:
                    logging.error(f"Failed to assign a GPU for {design_step} {index}. GPUs: {self.gpus}. Error in run_design() / main_design.py")
                    sys.exit()    

                # IS THIS THE PROBLEM???
                #gpu_summary = {key: ('busy' if value is not None else None) for key, value in self.gpus.items()}
                #gpu_summary[gpu_id] = 'assigned'
                logging.debug(f"Assigned GPU for {index}_{design_step}. GPUs: {self.gpus}")
        
        if design_step == "ProteinMPNN":
            cmd = prepare_ProteinMPNN(self, index, cmd, gpu_id = gpu_id)
            logging.info(f"Run ProteinMPNN for index {index}.")
        
        elif design_step == "SolubleMPNN":
            cmd = prepare_SolubleMPNN(self, index, cmd, gpu_id = gpu_id)
            logging.info(f"Run SolubleMPNN for index {index}.")

        elif design_step == "LigandMPNN":
            cmd = prepare_LigandMPNN(self, index, cmd, gpu_id = gpu_id)
            logging.info(f"Run LigandMPNN for index {index}.")
        
        elif design_step == "AlphaFold3MSA":
            cmd = prepare_AlphaFold3_MSA(self, index, cmd)
            logging.info(f"Run AlphaFold3MSA for index {index}.")

        elif design_step == "AlphaFold3INF":
            cmd = prepare_AlphaFold3_INF(self, index, cmd, gpu_id = gpu_id)
            logging.info(f"Run AlphaFold3INF for index {index}.")
            
        elif design_step == "RosettaDesign":
            cmd = prepare_RosettaDesign(self, index, cmd)
            logging.info(f"Run RosettaDesign for index {index} based on index {index}.")
            
        elif design_step == "RosettaRelax":
            cmd = prepare_RosettaRelax(self, index, cmd)
            logging.info(f"Run RosettaRelax for index {index}.")
        
        elif design_step == "MDMin":
            cmd = prepare_MDMin(self, index, cmd)
            logging.info(f"Run MD minimise for index {index}.")
            
        elif design_step == "ESMfold":
            cmd = prepare_ESMfold(self, index, cmd, gpu_id = gpu_id)
            logging.info(f"Run ESMfold for index {index}.")
            
        elif design_step == "ElectricFields":
            cmd = prepare_efields(self, index, cmd)
            logging.info(f"Calculating ElectricFields for index {index}.")
                        
        elif design_step == "BioDC":
            cmd = prepare_BioDC(self, index, cmd)
            logging.info(f"Calculating Redoxpotentials for index {index}.")
            
        else:
            logging.error(f"{design_step} is not defined! Error in run_design() / main_design.py")
            sys.exit()
                 
    # Write the shell command to a file and submit job                
    job = "_".join(design_steps)
    with open(f'{self.FOLDER_DESIGN}/{index}/scripts/{job}_{index}.sh','w') as file: file.write(cmd)
    submit_job(self, index=index, job=job, ram=ram, bash=bash)